import logging
import uuid
import sys
import matplotlib

from base64 import b16encode

import plotly.colors as clrs

from matplotlib import cm

import matplotlib.colors as mcolors

import numpy as np

class Options(object):
    def __init__(self,**kwargs):
        self._opts = {}

        for name,value in kwargs.items():
            self.setOpt(name,value)

    def hasOpt(self,name):
        code = False

        if isinstance(name,list):
            for n in name:
                if n.lower() in self._opts:
                    code = True
                    break
        else:
            code = name.lower() in self._opts

        return(code)

    def getOpt(self,name,dv = None,clear = False):
        value = None
        s = name.lower()

        if s in self._opts:
            value = self._opts[s]
            if clear:
                self.clear(s)

        if value == None and dv != None:
            value = dv
        
        return(value)

    def setOpt(self,name,value):
        s = name.lower()
        if value == None:
            if s in self._opts:
                del self._opts[s]
                self.optionSet(name,None)
        else:
            self._opts[s] = value
            self.optionSet(s,value)

    def optionSet(self,name,value):
        pass

    def clear(self,name):
        s = name.lower()
        if s in self._opts:
            del self._opts[s]

    def setOpts(self,**kwargs):
        for name,value in kwargs.items():
            self.setOpt(name,value)

    def items(self):
        return(self._opts.items())

    @property
    def options(self):
        return(self._opts)

    @options.setter
    def options(self,options):
        for name,value in six.iteritems(options):
            self.setOpt(name,value)

class Gradient(Options):
    def __init__(self,color,**kwargs):
        Options.__init__(self,**kwargs)

        c = Colors.getColorFromName(color)

        if c == None:
            raise Exception("invalid color: " + str(color))

        if len(c) != 7:
            raise Exception("invalid color: " + str(color))

        self._color = c
        self._levels = self.getOpt("levels",100)

        minv = self.getOpt("min",0)
        maxv = self.getOpt("max",100)

        self._a = []

        if maxv > minv:
            self._a = np.arange(minv,maxv,(maxv - minv) / self._levels)

    def darken(self,value):

        s = self._color

        if len(self._a) > 0:
            a = np.where(value >= self._a)[0]
            level = len(a) - 1

            rgbHex = [self._color[x:x + 2] for x in [1, 3, 5]]
            rgb = [int(v, 16) - level for v in rgbHex]
            rgb = [min([255, max([0,i])]) for i in rgb]

            s = "#{0:02x}{1:02x}{2:02x}".format(rgb[0],rgb[1],rgb[2])

        return(s)

    def brighten(self,value):

        s = self._color

        if len(self._a) > 0:
            a = np.where(value >= self._a)[0]
            level = len(a) - 1

            rgbHex = [self._color[x:x + 2] for x in [1, 3, 5]]
            rgb = [int(value, 16) + level for value in rgbHex]
            rgb = [min([255, max([0,i])]) for i in rgb]

            s = "#{0:02x}{1:02x}{2:02x}".format(rgb[0],rgb[1],rgb[2])

        return(s)

    @property
    def color(self):
        return(self._color)

    @color.setter
    def color(self,value):
        self._color = value

class Colors(Options):
    _sasThemes = {
        "sas_base":["#00929f", "#f08000", "#90b328", "#3d5aae", "#ffca39", "#a6427c", "#9c2910", "#736519"],
        "sas_dark":["#90b328", "#9c2910", "#ffca39", "#00929f", "#736519", "#f08000", "#a6427c"],
        "sas_highcontrast":["#a1d73b", "#ff791d", "#ffd736", "#cb66ff", "#ff5252", "#57b2ff", "#fa96e0", "#33f7b0"],
        "sas_light":["#3d5aae", "#90b328", "#9c2910", "#ffca39", "#00929f", "#736519", "#f08000", "#a6427c"],
        "sas_marine":["#00929f", "#f08000", "#90b328", "#3d5aae", "#ffca39", "#a6427c", "#9c2910", "#736519"],
        "sas_midnight":["#2470ad", "#98863c", "#5954ad", "#985b30", "#238a92", "#84414b", "#17785f", "#985186"],
        "sas_opal":["#33a3ff", "#ffcc32", "#9471ff", "#ff8224", "#2ad1d1", "#dd5757", "#15b57b", "#ff6fbd"],
        "sas_sail":["#21b9b7", "#4141e0", "#7db71a", "#8e2f8a", "#d38506", "#0abf85", "#2f90ec", "#db3851"],
        "sas_snow":["#3d5aae", "#90b328", "#9c2910", "#ffca39", "#00929f", "#736519", "#f08000", "#a6427c"],
        "sas_umstead":["#00929f", "#f08000", "#90b328", "#3d5aae", "#ffca39", "#a6427c", "#9c2910", "#736519"],
        "sas_corporate":["#00929f", "#f08000", "#90b328", "#3d5aae", "#ffca39", "#a6427c", "#9c2910", "#736519"],
        "sas_hcb":["#7cbf00", "#f77107", "#f1d700", "#bd77ff", "#ff6d65", "#4aacff", "#ff6fbd", "#00d692"],
        "sas_ignite":["#2470ad", "#98863c", "#5954ad", "#985b30", "#238a92", "#84414b", "#17785f", "#985186"],
        "sas_inspire":["#21b9b7", "#4141e0", "#7db71a", "#8e2f8a", "#d38506", "#0abf85", "#2f90ec", "#db3851"]
    }

    @staticmethod
    def getColorFromName(name):
        if name.find("#") == 0:
            return(name)

        colors = mcolors.get_named_colors_mapping()

        color = None

        if name in colors:
            color = colors[name]

        return(color)

    @staticmethod
    def getLuma(name):
        luma = None
        color = Colors.getColorFromName(name)
        if color != None:
            r = int(color[1:3],16)
            g = int(color[3:5],16)
            b = int(color[5:7],16)

            luma = 0.2126 * r + 0.7152 * g + 0.0722 * b

        return(luma)

    def __init__(self,**kwargs):
        Options.__init__(self,**kwargs)

        if self.hasOpt("colormap"):
            self.createFromColorMap(self.getOpt("colormap"))
        elif self.hasOpt("colors"):
            self.createFromColors(self.getOpt("colors"))

    def createFromColorMap(self,colormap):
        colors = []
        colorscale = []
        luma = []

        if colormap != None:
            if colormap.find("sas_") == 0:
                if colormap in Colors._sasThemes:
                    colors.extend(Colors._sasThemes[colormap])
            elif colormap in clrs.PLOTLY_SCALES:
                cmap = clrs.PLOTLY_SCALES[colormap]
                interval = 1 / (len(cmap) - 1)
                index = 0
                for i,c in enumerate(cmap):
                    s = c[1]
                    if s[0] == '#':
                        colors.append(s)
                    else:
                        i1 = s.index("(")
                        i2 = s.index(")")
                        s = s[i1 + 1:i2]
                        colorscale.append([index,"rgb(" + s + ")"])
                        a = s.split(",")
                        r = int(a[0])
                        g = int(a[1])
                        b = int(a[2])
                        value = (r,g,b)
                        luma.append(0.2126 * r + 0.7152 * g + 0.0722 * b)
                        s = "#" + b16encode(bytes(value)).decode()
                        colors.append(s)

                    if i == (len(cmap) - 2):
                        index = 1
                    else:
                        index += interval
            else:
                try:
                    cmap = matplotlib.cm.get_cmap(colormap)
                    norm = matplotlib.colors.Normalize(vmin = 0,vmax = 255)
                    rgb = []

                    for i in range(0, 255):
                        k = matplotlib.colors.colorConverter.to_rgb(cmap(norm(i)))
                        rgb.append(k)

                    entries = 255

                    h = 1.0 / (entries - 1)

                    prev = None

                    a = []

                    for i in range(entries):
                        c = list(map(np.uint8,np.array(cmap(i * h)[:3]) * 255))
                        value = (c[0],c[1],c[2])
                        if value == prev:
                            continue
                        luma.append(0.2126 * c[0] + 0.7152 * c[1] + 0.0722 * c[2])
                        prev = value
                        a.append(["#" + b16encode(bytes(value)).decode(),"rgb(" + str(c[0]) + "," + str(c[1]) + "," + str(c[2]) + ")"])

                    if len(a) > 1:
                        interval = 1 / (len(a) - 1)
                        index = 0

                        for i,x in enumerate(a):
                            colors.append(x[0])
                            colorscale.append([index,x[1]])
                            if i == (len(a) - 2):
                                index = 1
                            else:
                                index += interval

                except:
                    pass

        if len(colors) == 0:
            interval = 1 / (len(clrs.DEFAULT_PLOTLY_COLORS) - 1)
            index = 0
            for i,c in enumerate(clrs.DEFAULT_PLOTLY_COLORS):
                i1 = c.index("(")
                i2 = c.index(")")
                s = c[i1 + 1:i2]
                colorscale.append([index,"rgb(" + s + ")"])
                a = s.split(",")
                r = int(a[0])
                g = int(a[1])
                b = int(a[2])
                luma.append(0.2126 * r + 0.7152 * g + 0.0722 * b)
                value = (r,g,b)
                colors.append("#" + b16encode(bytes(value)).decode())

                if i == (len(clrs.DEFAULT_PLOTLY_COLORS) - 2):
                    index = 1
                else:
                    index += interval
        elif len(colorscale) == 0:
            interval = 1 / (len(colors) - 1)
            index = 0
            for i,c in enumerate(colors):
                r = int(c[1:3],16)
                g = int(c[3:5],16)
                b = int(c[5:7],16)

                colorscale.append([index,"rgb(" + str(r) + "," + str(g) + "," + str(b) + ")"])
                luma.append(0.2126 * r + 0.7152 * g + 0.0722 * b)

                if i == (len(colors) - 2):
                    index = 1
                else:
                    index += interval

        self._colors = colors
        self._colorscale = colorscale
        self._luma = luma 

    def createFromColors(self,colors):
        if len(colors) < 2:
            raise Exception("must have at least 2 colors")
            
        colorscale = []
        luma = []

        interval = 1 / (len(colors) - 1)
        index = 0

        for i,c in enumerate(colors):
            r = int(c[1:3],16)
            g = int(c[3:5],16)
            b = int(c[5:7],16)

            colorscale.append([index,"rgb(" + str(r) + "," + str(g) + "," + str(b) + ")"])
            luma.append(0.2126 * r + 0.7152 * g + 0.0722 * b)

            if i == (len(colors) - 2):
                index = 1
            else:
                index += interval

        self._colors = []
        self._colors.extend(colors)

        self._colorscale = colorscale
        self._luma = luma 

    def getColor(self,name):
        if name.find("#") == 0:
            return(name)

        colors = mcolors.get_named_colors_mapping()

        color = None

        if name in colors:
            color = colors[name]
        elif name == "lightest":
            color = self.lightest
        elif name == "darkest":
            color = self.darkest

        return(color)

    def getColors(self,num,increment):
        index = 0
        colors = []

        for i in range(0,num):
            colors.append(self._colors[index])
            index += increment
            if index == len(self._colors):
                index = 0

        return(colors)

    def getSpread(self,num):
        delta = int(len(self._colors) / num)
        return(self.getColors(num,delta))

    def getFirst(self,num = 1):
        colors = []

        index = 0

        for i in range(0,num):
            colors.append(self._colors[index])
            index += 1
            if index == len(self._colors):
                index = 0

        return(colors)

    def getClosestTo(self,luma):
        color = None
        if len(self._colors) > 0:
            index = -1
            diff = sys.maxsize
            for i,l in enumerate(self._luma):
                d = abs(luma - l)
                if d < diff:
                    diff = d
                    index = i

            if index >= 0:
                color = self._colors[index]

        return(color)

    @property
    def size(self):
        return(len(self._colors))

    @property
    def colors(self):
        return(self._colors)

    @property
    def colorscale(self):
        return(self._colorscale)

    @property
    def first(self):
        color = None
        if len(self._colors) > 0:
            color = self._colors[0]
        return(color)

    @property
    def last(self):
        color = None
        if len(self._colors) > 0:
            color = self._colors[len(self._colors) - 1]
        return(color)

    @property
    def lightest(self):
        color = None
        if len(self._colors) > 0:
            maxLuma = 0
            index = -1
            for i,l in enumerate(self._luma):
                if l > maxLuma:
                    maxLuma = l
                    index = i

            if index >= 0:
                color = self._colors[index]

        return(color)

    @property
    def darkest(self):
        color = None
        if len(self._colors) > 0:
            minLuma = sys.maxsize
            index = -1
            for i,l in enumerate(self._luma):
                if l < minLuma:
                    minLuma = l
                    index = i

            if index >= 0:
                color = self._colors[index]

        return(color)

class ColorRange(object):
    def __init__(self,colors,minv,maxv):
        self._colors = colors
        self._minv = minv
        self._maxv = maxv
        self._a = []
        if self._maxv > self._minv and len(self._colors.colors) > 0:
            self._a = np.arange(self._minv,self._maxv,(self._maxv - self._minv) / len(self._colors.colors))

    def getColor(self,value):
        color = None

        if len(self._a) > 0:
            index = np.where(value >= self._a)[0]
            color = self._colors.colors[len(index) - 1]
        elif len(self._colors.colors) > 0:
            color = self._colors.colors[0]
        else:
            color = "#ffffff"

        return(color)

def supports(o,method):
    code = False
    if (o != None):
        try:
            value = getattr(o,method)
            if (value != None):
                code = callable(value)
        except:
            pass
    return(code)

def contains(list,o):
    for item in list:
        if item == o:
            return(True)

    return(False)

def indexOf(list,o):
    i = 0
    for item in list:
        if item == o:
            return(i)
        i += 1

    return(-1)

def addTo(list,o):
    if contains(list,o) == False:
        list.append(o);
        return(True);

    return(False);

def removeFrom(list,o):
    code = False
    index = indexOf(list,o)
    if index >= 0:
        list.pop(index)
        code = True
    return(code)

def guid():
    return(str(uuid.uuid4()).replace('-', '_'))

def brighten(color,offset):
    if len(color) != 7:
        return("#ffffff")

    rgbHex = [color[x:x + 2] for x in [1, 3, 5]]
    rgb = [int(value, 16) + offset for value in rgbHex]
    rgb = [min([255, max([0,i])]) for i in rgb]

    s = "#{0:02x}{1:02x}{2:02x}".format(rgb[0],rgb[1],rgb[2])

    return(s)

def darken(color,offset):
    if len(color) != 7:
        return("#ffffff")

    rgbHex = [color[x:x + 2] for x in [1, 3, 5]]
    rgb = [int(value, 16) - offset for value in rgbHex]
    rgb = [min([255, max([0,i])]) for i in rgb]

    s = "#{0:02x}{1:02x}{2:02x}".format(rgb[0],rgb[1],rgb[2])

    return(s)

def convertColormap(name):
    cmap = matplotlib.cm.get_cmap(name)
    norm = matplotlib.colors.Normalize(vmin = 0,vmax = 255)
    rgb = []
 
    for i in range(0, 255):
        k = matplotlib.colors.colorConverter.to_rgb(cmap(norm(i)))
        rgb.append(k)

    entries = 255

    h = 1.0 / (entries - 1)
    colorscale = []

    for k in range(entries):
        C = list(map(np.uint8,np.array(cmap(k * h)[:3]) * 255))
        colorscale.append([k * h,"rgb" + str((C[0], C[1], C[2]))])

    return(colorscale)

def convertColormapToPalette(name):
    cmap = matplotlib.cm.get_cmap(name)
    norm = matplotlib.colors.Normalize(vmin = 0,vmax = 255)
    rgb = []
 
    for i in range(0, 255):
        k = matplotlib.colors.colorConverter.to_rgb(cmap(norm(i)))
        rgb.append(k)

    entries = 255

    h = 1.0 / (entries - 1)
    colorscale = []

    prev = None

    for k in range(entries):
        c = list(map(np.uint8,np.array(cmap(k * h)[:3]) * 255))
        value = (c[0],c[1],c[2])
        if value == prev:
            continue
        prev = value
        s = "#" + b16encode(bytes(value)).decode()
        colorscale.append(s)

    return(colorscale)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

import os
import pandas as pd
import six
from .base import BaseWindow, attribute
from .features import SchemaFeature, ModelsFeature, ConnectorsFeature
from .utils import get_args, ensure_element


class ScoreWindow(BaseWindow, SchemaFeature, ModelsFeature, ConnectorsFeature):
    '''
    Score window

    Parameters
    ----------
    name : string, optional
        The name of the window
    schema : Schema, optional
        The schema of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level
        value of pubsub is manual, true enables publishing and subscribing
        for the window and false disables it.
    description : string, optional
        Description of the window

    Attributes
    ----------
    online_models : list-of-OnlineModels
        List of online model objects
    offline_models : list-of-OfflineModels
        List of offline model objects

    Returns
    -------
    :class:`ScoreWindow`

    '''

    window_type = 'score'

    def __init__(self, name=None, schema=None, pubsub=None, description=None,
                 copyvars=None):
        BaseWindow.__init__(self, **get_args(locals()))
        # Set the online model for subclasses
        if type(self).__name__ != 'ScoreWindow':
            self.add_online_model(type(self).__name__)

    def _create_schema_list(self, variables):
        '''
        Extract schema information from DataFrame

        Parameters
        ----------
        variables : DataFrame
            The DataFrame containing schema information

        Returns
        -------
        list

        '''
        labels = []
        labels.append('id*:int64')
        for name, dtype in zip(variables['Name'], variables['Type']):
            if dtype == 'Num':
                labels.append(name + ':double')
            elif dtype == 'Char':
                labels.append(name + ':string')
        return labels

    def import_schema_from_astore_output(self, output_variables_input):
        '''
        Import a schema from the astore CAS action output format

        Parameters
        ----------
        output_variables_input : DataFrame or list or string
            The schema definition

        '''
        if isinstance(output_variables_input, six.string_types):
            if os.path.isfile(output_variables_input):
                output_variables_input = pd.read_csv(output_variables_input)
            else:
                output_variables_input = pd.read_csv(six.StringIO(output_variables_input))
        if isinstance(output_variables_input, pd.DataFrame):
            self.schema = self._create_schema_list(output_variables_input)
        elif isinstance(output_variables_input, (tuple, list)):
            self.schema = list(output_variables_input)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Kafka Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class KafkaSubscriber(Connector):
    '''
    Subscribe to events from a Kafka broker

    Parameters
    ----------
    kafkahostport : string
        Specifies one or more Kafka brokers in the following form:
        'host:port,host:port,...'.
    kafkatopic : string
        Specifies the Kafka topic
    urlhostport : string
        Specifies the host:port field in the metadata topic that is
        subscribed to on start-up.
    kafkapartition : string, optional
        Specifies the Kafka partition
    kafkatype : string, optional
        Specifies binary, CSV, JSON, or the name of a string field in
        the subscribed window schema.
    numbufferedmsgs : int, optional
        Specifies the maximum number of messages buffered by a
        standby subscriber connector.
    name : string, optional
        The name of the connector object
    snapshot : bool, optional
        Specifies whether to send snapshot data
    collapse : bool, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber
        output publishable.
    rmretdel : bool, optional
        Specifies to remove all delete events from event blocks received
        by a subscriber that were introduced by a window retention policy.
    hotfailover : bool, optional
        Enables hot failover mode
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields
        in CSV events.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol
        Buffers message definition used to convert event blocks to
        protobuf messages.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the
        .proto file that you specified with the protofile parameter.
    csvincludeschema : string, optional
        When kafkatype=CSV, specifies when to prepend output CSV data
        with the window's serialized schema.
        Valid values: 'never', 'once', and 'pereventblock'
    useclientmsgid : bool, optional
        Uses the client-generated message ID instead of the engine-generated
        message ID when performing a failover operation and extracting a
        message ID from an event block.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters.
    zookeeperhostport : string, optional
        Specifies the Zookeeper server in the form 'host:port'
    kafkaglobalconfig : string, optional
        Specifies a semicolon-separated list of 'key=value' strings to
        configure librdkafka global configuration values.
    kafkatopicconfig : string, optional
        Specifies a semicolon-separated list of 'key=value' strings to
        configure librdkafka topic configuration values.
    csvmsgperevent : bool, optional
        For CSV, specifies to send one message per event
    csvmsgperevent_block : bool, optional
        For CSV, specifies to send one message per event block

    Returns
    -------
    :class:`KafkaSubscriber`

    '''
    connector_key = dict(cls='kafka', type='subscribe')

    property_defs = dict(
        kafkahostport=prop('kafkahostport', dtype='string', required=True,
                           valid_values=re.compile(r'(\w[\w\-\.]*:\d+\s*,?\s*)+')),
        kafkatopic=prop('kafkatopic', dtype='string', required=True),
        kafkapartition=prop('kafkapartition', dtype='string', required=True, default=0),
        kafkatype=prop('kafkatype', dtype='string', required=True, default='csv'),
        urlhostport=prop('urlhostport', dtype='string', required=True),
        numbufferedmsgs=prop('numbufferedmsgs', dtype='int', required=True,
                             default=10000, valid_expr='value >= 0'),
        snapshot=prop('snapshot', dtype='bool', required=True, default=False),
        collapse=prop('collapse', dtype='bool'),
        rmretdel=prop('rmretdel', dtype='bool'),
        hotfailover=prop('hotfailover', dtype='bool'),
        dateformat=prop('dateformat', dtype='string'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        csvincludeschema=prop('csvincludeschema', dtype='string',
                              valid_values=['never', 'once', 'pereventblock']),
        useclientmsgid=prop('useclientmsgid', dtype='bool'),
        configfilesection=prop('configfilesection', dtype='string'),
        zookeeperhostport=prop('zookeeperhostport', dtype='string',
                               valid_values=re.compile(r'\w[\w\-\.]*:\d+')),
        kafkaglobalconfig=prop('kafkaglobalconfig', dtype='string'),
        kafkatopicconfig=prop('kafkatopicconfig', dtype='string'),
        csvmsgperevent=prop('csvmsgperevent', dtype='bool'),
        csvmsgperevent_block=prop('csvmsgpereventblock', dtype='bool'),
    )

    def __init__(self, kafkahostport=None, kafkatopic=None, urlhostport=None,
                 kafkapartition=None, kafkatype=None, numbufferedmsgs=None,
                 name=None, is_active=None, snapshot=None, collapse=None,
                 rmretdel=None, hotfailover=None, dateformat=None,
                 protofile=None, protomsg=None,
                 csvincludeschema=None, useclientmsgid=None,
                 configfilesection=None, zookeeperhostport=None,
                 kafkaglobalconfig=None, kafkatopicconfig=None,
                 csvmsgperevent=None, csvmsgpereventblock=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'kafka', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['kafkahostport',
                                                   'kafkatopic',
                                                   'urlhostport'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)


class KafkaPublisher(Connector):
    '''
    Publish events to a Kafka broker

    Parameters
    ----------
    kafkahostport : string
        Specifies one or more Kafka brokers in the following form:
        'host:port,host:port,...'.
    kafkatopic : string
        Specifies the Kafka topic
    urlhostport : string
        Specifies the host:port field in the metadata topic that is
        subscribed to on start-up.
    kafkapartition : string, optional
        Specifies the Kafka partition
    kafkatype : string, optional
        Specifies binary, CSV, JSON, or the name of a string field in
        the subscribed window schema.
    name : string, optional
        The name of the connector object
    transactional : string, optional
        When kafkatype = CSV, sets the event block type to transactional.
        The default value is normal.
    blocksize : int, optional
        When kafkatype = CSV, specifies the number of events to include in
        a published event block. The default value is 1.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields
        in CSV events.
    ignorecsvparseerrors : boolean, optional
        Specifies that when a field in an input CSV event cannot be parsed,
        the event is dropped, an error is logged, and publishing continues.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol Buffers
        message definition used to convert event blocks to protobuf messages.
        When you specify this parameter, you must also specify the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the .proto
        file that you specified with the protofile parameter. Event blocks
        are converted into this message.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters.
    csvfielddelimiter : string, optional
        Specifies the character delimiter for field data in input CSV
        events. The default delimiter is the ',' character.
    noautogenfield : boolean, optional
        Specifies that input events are missing the key field that is
        autogenerated by the source window.
    publishwithupsert : boolean, optional
        Specifies to build events with opcode=Upsert instead of opcode=Insert.
    kafkainitialoffset : string or int, optional
        Specifies the offset from which to begin consuming messages from the
        Kafka topic and partition. Valid values are "smallest", "largest",
        or an integer. The default value is "smallest".
    addcsvopcode : boolean, optional
        Prepends an opcode and comma to write CSV events. The opcode is
        Insert unless publish_with_upsert is enabled.
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV events (with a comma).
        Valid values are "normal" and "partialupdate".
    kafkaglobalconfig : string, optional
        Specifies a semicolon-separated list of "key=value" strings to configure
        librdkafka global configuration values.
    kafkatopicconfig : string, optional
        Specifies a semicolon-separated list of "key=value" strings to configure
        librdkafka topic configuration values.
    useclientmsgid : boolean, optional
        If the Source window has been restored from a persist to disk, ignore
        received binary event blocks that contain a message ID less than the
        greatest message ID in the restored window.
    maxevents : int, optional
         Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`KafkaPublisher`

    '''
    connector_key = dict(cls='kafka', type='publish')

    property_defs = dict(
        kafkahostport=prop('kafkahostport', dtype='string', required=True,
                           valid_values=re.compile(r'(.+?:.+?\s*,?\s*)+')),
        kafkatopic=prop('kafkatopic', dtype='string', required=True),
        kafkapartition=prop('kafkapartition', dtype='string', required=True),
        kafkatype=prop('kafkatype', dtype='string', required=True),
        urlhostport=prop('urlhostport', dtype='string', required=True),
        name=prop('name', dtype='string'),
        transactional=prop('transactional', dtype='string'),
        blocksize=prop('blocksize', dtype='int'),
        dateformat=prop('dateformat', dtype='string'),
        ignorecsvparseerrors=prop('ignorecsvparseerrors', dtype='boolean'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        csvfielddelimiter=prop('csvfielddelimiter', dtype='string'),
        noautogenfield=prop('noautogenfield', dtype='boolean'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        kafkainitialoffset=prop('kafkainitialoffset', dtype=('string', 'int')),
        addcsvopcode=prop('addcsvopcode', dtype='boolean'),
        addcsvflags=prop('addcsvflags', dtype='boolean'),
        kafkaglobalconfig=prop('kafkaglobalconfig', dtype='string'),
        kafkatopicconfig=prop('kafkatopicconfig', dtype='string'),
        useclientmsgid=prop('useclientmsgid', dtype='boolean'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, kafkahostport=None, kafkatopic=None, urlhostport=None,
                 kafkapartition=None, kafkatype=None, name=None, is_active=None,
                 transactional=None, blocksize=None, dateformat=None,
                 ignorecsvparseerrors=None, protofile=None, protomsg=None,
                 configfilesection=None, csvfielddelimiter=None,
                 noautogenfield=None, publishwithupsert=None,
                 kafkainitialoffset=None, addcsvopcode=None,
                 addcsvflags=None, kafkaglobalconfig=None,
                 kafkatopicconfig=None, useclientmsgid=None,
                 maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'kafka', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['kafkahostport',
                                                   'kafkatopic',
                                                   'urlhostport'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from xml.etree import ElementTree as xml
from .base import Window, BaseWindow, attribute
from .utils import get_args

class RemoveStateWindow(Window):
    '''
    RemoveStateWindow Window

    Parameters
    ----------
    name : string, optional
        The name of the window
    description : string, optional
        Description of the window
    pubsub : bool, optional
        Publish/subscribe mode for the window. When the project-level value
        of pubsub is manual, true enables publishing and subscribing for
        the window and false disables it.
   add_log_fields : bool, optional
        Add originalOC, originalFL to produced events, default=false
   remove : string, optional
        Optional list filtered out attributes

    Returns
    -------
    :class:`RemoveStateWindow`

    '''

    window_type = 'remove-state'

    remove = attribute('remove',dtype='string')
    add_log_fields = attribute('add-log-fields',dtype='string')

    def __init__(self,remove=None,add_log_fields=False,
                 name=None,description=None,pubsub=False):
        Window.__init__(self, **get_args(locals()))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import os
import six
import re
from.template import Template

import logging

logging.basicConfig(filename=os.getenv("ESPPY_LOG"), level=logging.INFO)


def builtin_template_helper(cls, template_name, file_name,  **kwargs):
    '''
    Helper function during initialization
    '''

    out = cls.from_xml(os.path.join(cls.file_dir, 'xmls', file_name), template_name)

    for key, value in six.iteritems(kwargs):
        for each in out.required_parameter_map.get(key, key):
            window, actual_key = re.split('\.', each)
            single_args = dict({actual_key: value})
            if actual_key == 'input_map':
                out.set_inputs(window, **single_args)
            elif actual_key == 'output_map':
                out.set_outputs(window, **single_args)
            else:
                out.set_parameters(window, **single_args)

    return out


def add_method(cls, file_name):
    ky = file_name.split('.xml')[0]

    temp = cls.from_xml(os.path.join(cls.file_dir, 'xmls', file_name), None)
    cls.template_list[ky]['description'] = temp.description
    cls.template_list[ky]['required_parameter_map'] = temp.required_parameter_map

    def builtin_template(template_name, **kwargs):
        '''
        Create a %s template

        Parameters
        ----------
        template_name : string
            the name of template

        Returns
        -------
        :class:`Template`

        '''
        return builtin_template_helper(cls, template_name, file_name, **kwargs)

    builtin_template.__name__ = ky
    builtin_template.__doc__ %= builtin_template.__name__
    setattr(cls, builtin_template.__name__, builtin_template)


for builtin_file in os.listdir(os.path.join(Template.file_dir, "xmls")):
    if builtin_file.endswith(".xml"):
        add_method(Template, builtin_file)





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Project Utilities '''

from __future__ import print_function, division, absolute_import, unicode_literals


def expand_path(path):
    '''
    Convert string path to window path components

    Parameters
    ----------
    path : string
        A '.' or '/' delimited string containing a project,
        continuous query, or window path.

    Returns
    -------
    list of list-of-strings or strings or None

    '''
    parts = None

    if path is None:
        return []

    if isinstance(path, (list, tuple)):
        path = '|'.join(path)

    if '.' in path:
        parts = [x.split('|') for x in path.split('.')]
    elif '/' in path:
        parts = [x.split('|') for x in path.split('/')]
    else:
        parts = [path.split('|')]

    for i, part in enumerate(parts):
        if '*' in part:
            parts[i] = None
        elif len(part) == 1:
            parts[i] = part[0]

    return parts



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Tibco Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class TibcoSubscriber(Connector):
    '''
    Subscribe to Tibco Rendezvous (RV) events

    Parameters
    ----------
    tibrvsubject : string
        Specifies the Tibco RV subject name
    tibrvtype : string
        Specifies binary, CSV, JSON, or the name of a string field in
        the subscribed window schema.
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    collapse : string, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber
        output publishable. The default value is disabled.
    tibrvservice : string, optional
        Specifies the Rendezvous service used by the Tibco RV transport
        created by the connector. The default service name is “rendezvous”.
    tibrvnetwork : string, optional
        Specifies the network interface used by the Tibco RV transport
        created by the connector. The default network depends on the
        type of daemon used by the connector.
    tibrvdaemon : string, optional
        Specifies the Rendezvous daemon used by the connector. The
        default is the default socket created by the local daemon.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks received
        by a subscriber that were introduced by a window retention policy.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields in
        CSV events. The default behavior is these fields are interpreted
        as an integer number of seconds (ESP_DATETIME) or microseconds
        (ESP_TIMESTAMP) since epoch.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol Buffers
        message definition. This definition is used to convert event
        blocks to protobuf messages. When you specify this parameter,
        you must also specify the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the
        .proto file that you specified with the protofile parameter.
        Event blocks are converted into this message.
    csvmsgperevent : int, optional
        For CSV, specifies to send one message per event. The default is
        one message per transactional event block or else one message
        per event.
    csvmsgpereventblock : int, optional
        For CSV, specifies to send one message per event block. The
        default is one message per transactional event block or else one
        message per event.

    Returns
    -------
    :class:`TibcoSubscriber`

    '''
    connector_key = dict(cls='tibrv', type='subscribe')

    property_defs = dict(
        tibrvsubject=prop('tibrvsubject', dtype='string', required=True),
        tibrvtype=prop('tibrvtype', dtype='string', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        collapse=prop('collapse', dtype='string'),
        tibrvservice=prop('tibrvservice', dtype='string'),
        tibrvnetwork=prop('tibrvnetwork', dtype='string'),
        tibrvdaemon=prop('tibrvdaemon', dtype='string'),
        rmretdel=prop('rmretdel', dtype='boolean'),
        dateformat=prop('dateformat', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        csvmsgperevent=prop('csvmsgperevent', dtype='int'),
        csvmsgpereventblock=prop('csvmsgpereventblock', dtype='int')
    )

    def __init__(self, tibrvsubject=None, tibrvtype=None,
                 name=None, is_active=None,
                 snapshot=None, collapse=None, tibrvservice=None,
                 tibrvnetwork=None, tibrvdaemon=None, rmretdel=None,
                 dateformat=None, configfilesection=None,
                 protofile=None, protomsg=None, csvmsgperevent=None,
                 csvmsgpereventblock=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'tibrv', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['tibrvsubject',
                                                   'tibrvtype'],
                                         delete='type')
        return cls(req[0], req[1], name=name, is_active=is_active, **properties)


class TibcoPublisher(Connector):
    '''
    Subscribe to Tibco Rendezvous (RV) events

    Parameters
    ----------
    tibrvsubject : string
        Specifies the Tibco RV subject name
    tibrvtype : string
        Specifies binary, CSV, JSON, or opaquestring. For opaquestring,
        the Source window schema is assumed to be "index:int64,message:string".
    blocksize : int, optional
        Specifies the number of events to include in a published
        event block. The default value is 1.
    transactional : string, optional
        Sets the event block type to transactional. The default value
        is normal.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields
        in CSV events. The default behavior is these fields are
        interpreted as an integer number of seconds (ESP_DATETIME)
        or microseconds (ESP_TIMESTAMP) since epoch.
    tibrvservice : string, optional
        Specifies the Rendezvous service used by the Tibco RV
        transport created by the connector. The default service
        name is “rendezvous”.
    tibrvnetwork : string, optional
        Specifies the network interface used by the Tibco RV transport
        created by the connector. The default network depends
        on the type of daemon used by the connector.
    tibrvdaemon : string, optional
        Specifies the Rendezvous daemon used by the connector. The
        default is the default socket created by the local daemon.
    configfilesection : string, optional
        Specifies the name of the section in the connector config
        file to parse for configuration parameters. Specify the
        value as [configfilesection].
    ignorecsvparseerrors : boolean, optional
        Specifies that when a field in an input CSV event cannot be
        parsed, the event is dropped, an error is logged, and
        publishing continues.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol
        Buffers message definition. This definition is used to
        convert event blocks to protobuf messages. When you specify
        this parameter, you must also specify the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in
        the .proto file that you specified with the protofile
        parameter. Event blocks are converted into this message.
    csvfielddelimiter : string, optional
        Specifies the character delimiter for field data in input
        CSV events. The default delimiter is the , character.
    noautogenfield : boolean, optional
        Specifies that input events are missing the key field that
        is autogenerated by the source window.
    publishwithupsert : boolean, optional
        Specifies to build events with opcode = Upsert instead of Insert.
    addcsvopcode : boolean, optioanl
        Prepends an opcode and comma to input CSV events. The
        opcode is Insert unless publishwithupsert is enabled.
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV events
        (with a comma). Valid values are "normal" and "partialupdate".
    maxevents : int, optional
        Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`TibcoPublisher`

    '''
    connector_key = dict(cls='tibrv', type='publish')

    property_defs = dict(
        tibrvsubject=prop('tibrvsubject', dtype='string', required=True),
        tibrvtype=prop('tibrvtype', dtype='string', required=True),
        blocksize=prop('blocksize', dtype='int'),
        transactional=prop('transactional', dtype='string'),
        dateformat=prop('dateformat', dtype='string'),
        tibrvservice=prop('tibrvservice', dtype='string'),
        tibrvnetwork=prop('tibrvnetwork', dtype='string'),
        tibrvdaemon=prop('tibrvdaemon', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        ignorecsvparseerrors=prop('ignorecsvparseerrors', dtype='boolean'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        csvfielddelimiter=prop('csvfielddelimiter', dtype='string'),
        noautogenfield=prop('noautogenfield', dtype='boolean'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        addcsvopcode=prop('addcsvopcode', dtype='boolean'),
        addcsvflags=prop('addcsvflags', dtype='string'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, tibrvsubject=None, tibrvtype=None,
                 name=None, is_active=None,
                 blocksize=None, transactional=None, dateformat=None,
                 tibrvservice=None, tibrvnetwork=None, tibrvdaemon=None,
                 configfilesection=None, ignorecsvparseerrors=None,
                 protofile=None, protomsg=None, csvfielddelimiter=None,
                 noautogenfield=None, publishwithupsert=None,
                 addcsvopcode=None, addcsvflags=None, maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'tibrv', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['tibrvsubject',
                                                   'tibrvtype'],
                                         delete='type')
        return cls(req[0], req[1], name=name, is_active=is_active, **properties)



import abc
import inspect
import ntpath
from keyword import iskeyword
import re
import six
import textwrap


def _is_valid_name(name):
    '''
    check if a string is a valid identifier in Python
    '''
    return name.isidentifier() and not iskeyword(name)


def _to_valid_name(varStr):
    return re.sub('\W|^(?=\d)', '_', varStr)


def _to_valid_dict(old_dict):
    '''
    convert the keys in a dict to valid identifiers

    Parameters:
    -------------
    old_dict: dict
        a dict whose keys may not be valid Python identifiers

    Returns
    -------------
    new_dict: dict
        a dict with all keys being valid Python identifiers
    map_dict: dict
        a dict recording the convert
        key: converted key; value: original key

    Examples:
    -------------
    old_key = {'a b':'float', 'c':'str'}
    new_dict = {'a_b': 'float', 'c':'str'}
    map_dict = {'a_b':'a b'}

    '''
    # dict to hold the valid variable name and its corrsponding datatype
    new_dict = old_dict.copy()
    # dict to hold the valid variable name and its corrsponding non-valid name
    map_dict = dict()
    for key in old_dict:
        # check if key is a valid variable name
        if not _is_valid_name(key):
            # make a valid name
            new_dict[_to_valid_name(key)] = new_dict.pop(key)
            # record it in map_dict
            map_dict[_to_valid_name(key)] = key
    return new_dict, map_dict


class wrap_generator(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def gen_wrap_str(self):
        raise NotImplementedError("gen_wrap_str must be defined!")

    @property
    @abc.abstractmethod
    def type(self):
        raise NotImplementedError(
            "the type of the generator must be specified!")


class KS_generator(wrap_generator):
    """
    Generator to generate wrapper code for exported Keras models.

    Parameters
    -----------
    h5_file : string
        The path to hdf5 file that stores the model structure and parameters.
        ESP server should be able to find this file.
    input_name : string
        Name of input array (features).
        This name should match the schema of the source window
    output_name : string
        Name of output (predictions).
    output_class : bool
        If True, the output is the predicted class. If False, the output is
        an array of pedicted probabilities of each class.
        Only applicable to classification models.

    """
    type = 'keras'

    def __init__(self, h5_file, input_name='input', output_name='output', output_class=False):
        self.file = h5_file
        self.input_name = input_name
        self.output_name = output_name
        self.output_class = output_class

    def gen_wrap_str(self):
        if self.output_class:
            predict = 'predict_classes'
        else:
            predict = 'predict'

        wrap_str = '''
model = None
def ks_score({0}):
    "Output: {1}"
    from keras.models import load_model
    import tensorflow as tf
    import numpy as np
    global model
    global graph
    # If it is called for the first time, restore the model
    if model is None:
        model = load_model('{2}')
        model._make_predict_function()
        graph = tf.get_default_graph()

    # make prediction
    {0}_wrap = np.array([{0}])
    with graph.as_default():
        {1} = model.{3}({0}_wrap)[0]

    if isinstance({1}, np.ndarray):
        {1} = {1}.tolist()
    else:
        {1} = {1}.item()

    return {1}'''.format(self.input_name, self.output_name, self.file, predict)
        return wrap_str


class TF_generator(wrap_generator):
    """
    Generator to generate wrapper code for exported Tensorflow models.

    Parameters
    -----------
    meta_file : string
        the path to meta file that stores the graph structure.
        The checkpoint files should be within the same directory with the meta file.
        ESP server should be able to find the files.
    input_op : string
        Name of input operation
    score_op : string
        Name of scoring operation
    input_name : string
        Name of input array (features).
        This name should match the schema of the source window
    output_name : string
        Name of output (predictions).

    Notes
    -----
    The Tensorflow models are expoted in checkpoint files using tf.train.Saver class.
    The name of input and scoring operations should be specified when creating the model.
    """
    type = 'tensorflow'

    def __init__(self, meta_file, input_op, score_op, input_name='input', output_name='output'):
        self.file = meta_file
        self.input_op = input_op
        self.score_op = score_op
        self.input_name = input_name
        self.output_name = output_name

    def gen_wrap_str(self):
        dir_path = ntpath.dirname(self.file) + '/'
        wrap_str = '''
sess = None
def tf_score({4}):
    "Output: {5}"
    import tensorflow as tf
    import numpy as np
    global sess
    global score_op
    global input_op
    #If it is called for the first time, restore the model and necessary operations
    if sess is None:
        sess=tf.Session()
        #load meta graph and restore weights
        saver = tf.train.import_meta_graph('{0}')
        saver.restore(sess,tf.train.latest_checkpoint('{1}'))

        graph = tf.get_default_graph()
        #restore the ops. Both ops were pre-defined in the model.
        input_op = graph.get_tensor_by_name("{2}:0") #op to feed input data
        score_op = graph.get_tensor_by_name("{3}:0")    #op to score the input

    #Note that the feed value of x has shape (?,xyz), NOT (,xyz)
    {4}_wrap = np.array([{4}])
    {5} = sess.run(score_op, feed_dict={{input_op: {4}_wrap}})[0]

    if isinstance({5}, np.ndarray):
        {5} = {5}.tolist()
    else:
        {5} = {5}.item()

    return {5}'''.format(self.file, dir_path, self.input_op, self.score_op,
                         self.input_name, self.output_name)

        return wrap_str


class JMP_generator(wrap_generator):
    """
    Add the infos of a JMP model

    Parameters
    -----------
    score_file : string
        The path to the model file exported by JMP
    """
    type = 'jmp'

    def __init__(self, score_file):
        file_string = open(score_file, 'r').read()
        file_string = file_string.replace('import jmp_score as jmp', '')
        file_string = file_string.replace('from __future__ import division', '')
        file_string = file_string.replace('from math import *', '')
        file_string = file_string.replace('import numpy as np', '')
        self.file_string = file_string
        six.exec_(self.file_string, globals())

        try:
            self.input_dict = getInputMetadata()
            self.output_dict = getOutputMetadata()
            self.model_dict = getModelMetadata()
        except NameError:
            raise ValueError("The file is not a valid Python scroing file from JMP")

    def _gen_schema(self, copy_var=None):
        # generate schema if not provided by user
        valid_output_dict, _ = _to_valid_dict(self.output_dict)
        type_dict = {'str': 'string', 'float': 'double'}
        schema = ['id*:int64']
        for key, value in valid_output_dict.items():
            for jmp_type, esp_type in type_dict.items():
                value = value.replace(jmp_type, esp_type)
            schema.append(key + ':' + value)

        if copy_var is not None:
            if isinstance(copy_var, six.string_types):
                schema.append(copy_var)
            elif isinstance(copy_var, (tuple, list)):
                schema = schema + list(copy_var)
        return schema

    def _set_jmp_funcs(self):
        from . import jmp_score
        func_source = inspect.getsource(jmp_score)
        # empty jmp class
        jmp_class = '''\
class jmp:
    pass\n\n'''
        # set each function in jmp_score as an attribute of jmp class
        set_attr = ''''''
        func_list = [o for o in inspect.getmembers(jmp_score, inspect.isfunction)]
        for func_name, _ in func_list:
            set_attr += '''setattr(jmp, "{0}", {0})\n'''.format(func_name)
        return func_source + jmp_class + set_attr

    def gen_wrap_str(self):
        valid_input_dict, input_map = _to_valid_dict(self.input_dict)
        valid_output_dict, output_map = _to_valid_dict(self.output_dict)

        intent = '    '
        # import necessary modules
        import_str = '''
from __future__ import division
import inspect
import numpy as np
from math import *\n\n'''

        jmp_funcs = self._set_jmp_funcs()
        jmp_class = '''
try:
    temp = jmp()
except NameError:
{}\n'''.format(textwrap.indent(jmp_funcs, 4 * ' '))

        # original score function
        orig_code = self.file_string + '\n\n'

        # create the signiture of the wrapper
        signiture = "def jmp_score(" + \
            ",".join(valid_input_dict.keys()) + '):\n'

        # create the docstring
        docstring = '''"Output: ''' + \
            ", ".join(valid_output_dict.keys()) + '''"'''

        # convert tuple to dict(indata)
        t_to_dict = '''
    frame = inspect.currentframe()
    _, _, _, values = inspect.getargvalues(frame)
    values.pop('frame', None)
    indata = values.copy()\n'''
        for key, value in input_map.items():
            t_to_dict += intent + \
                '''indata["''' + value + \
                '''"] = indata.pop("''' + key + '''")\n'''

        # score the input
        score_input = '''
    outdata=dict()
    score(indata, outdata)\n'''

        # convert outdata(dict)
        outdata_split = ''
        for key, value in output_map.items():
            outdata_split += intent + \
                '''outdata["''' + key + \
                '''"] = outdata.pop("''' + value + '''")\n'''
        for key in valid_output_dict:
            outdata_split += intent + key + \
                '''= outdata["''' + key + '''"]\n'''

        # return outputs
        return_outputs = '''return ''' + ",".join(valid_output_dict.keys())

        return import_str + jmp_class + orig_code + signiture + intent + docstring + \
            t_to_dict + score_input + outdata_split + intent + return_outputs



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import os
import unittest
import esppy
from . import utils as tm


class TestPatternWindow(tm.WindowTestCase):

    def test_xml(self):
        self._test_model_file('pattern_window.xml')

    def test_api(self):
        esp = self.s

        proj = esp.create_project('project_01_UnitTest', pubsub='auto',
                                  n_threads='1', compress_open_patterns=True)

        cq = esp.ContinuousQuery(name='cq_01')

        src = esp.SourceWindow(name='src_win', 
                               schema=('ID*:int32','symbol:string',
                                       'price:double','buy:int32','tradeTime:date'))
        src.add_connector('fs',
                          conn_name='pub', conn_type='publish',
                          type='pub',
                          fstype='csv',
                          fsname='pattern_window.csv',
                          transactional=True,
                          blocksize='1',
                          dateformat='%Y-%m-%d %H:%M:%S')

        pattern_win = esp.PatternWindow(name='pattern_win',
                                        schema=('ID*:int64','ID1:int32','ID2:int32'))
        pattern1=pattern_win.create_pattern(name='pattern1')
        pattern1.add_event('src_win','e1','symbol=="IBM" and price > 100.00 and b == buy')
        pattern1.add_event('src_win','e2','symbol=="SUN" and price > 25.000 and b == buy')
        pattern1.set_logic('fby(e1, e2)')
        pattern1.add_field_selection('e1','ID')
        pattern1.add_field_selection('e2','ID')
        pattern1.add_timefield('tradeTime','src_win')
        pattern_win.add_connector('fs',
                                  conn_name='sub', conn_type='subscribe',
                                  type='sub',
                                  fstype='csv',
                                  fsname='result.out',
                                  snapshot=True)

        proj.add_query(cq)
        cq.add_window(src)
        cq.add_window(pattern_win)

        src.add_target(pattern_win)

        self._test_model_file('pattern_window.xml', proj)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Teradata Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class TeradataSubscriber(Connector):
    '''
    Subscribe to Teradata operations

    Parameters
    ----------
    tdatatdpid : string
        Specifies the target Teradata server name
    desttablename : string
        Specifies the target table name
    tdatausername : string
        Specifies the user name for the user account on the target
        Teradata server.
    tdatauserpwd : string
        Specifies the user password for the user account on the target
        Teradata server.
    tdatamaxsessions : int
        Specifies the maximum number of sessions created by the TPT to
        the Teradata server.
    tdataminsessions : int
        Specifies the minimum number of sessions created by the TPT to
        the Teradata server.
    tdatadriver : string
        Specifies the operator: stream, update, or load.
    tdatainsertonly : boolean
        Specifies whether events in the subscriber event stream processing
        window are insert only. Must be true when using the load operator.
    snapshot : boolean
        Specifies whether to send snapshot data
    rmretdel : boolean, optional
        Removes all delete events from event blocks received by the
        subscriber that were introduced by a window retention policy.
    tdatabatchperiod : int, optional
        Specifies the batch period in seconds. Required when using the
        update operator, otherwise ignored.
    stage1tablename : string, optional
        Specifies the first staging table. Required when using the load
        operator, otherwise ignored.
    stage2tablename : string, optional
        Specifies the second staging table. Required when using the load
        operator, otherwise ignored.
    connectstring : string, optional
        Specifies the connect string used to access the target and
        staging tables. Use the form “DSN=dsnname;UID=userid;pwd=password”.
        Required when using the load operator, otherwise ignored.
    connectstring : string, optional
        Specifies the connect string used to access the target and
        staging tables. 
    tdatatracelevel : int, optional
        Specifies the trace level for Teradata messages written to the
        trace file in the current working directory.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    tdatauserpwdencrypted : boolean, optional
        Specifies that tdatauserpwd is encrypted

    Returns
    -------
    :class:`TeradataSubscriber`

    '''
    connector_key = dict(cls='tdata', type='subscribe')

    property_defs = dict(
        tdatatdpid=prop('tdatatdpid', dtype='string', required=True),
        desttablename=prop('desttablename', dtype='string', required=True),
        tdatausername=prop('tdatausername', dtype='string', required=True),
        tdatauserpwd=prop('tdatauserpwd', dtype='string', required=True),
        tdatamaxsessions=prop('tdatamaxsessions', dtype='int', required=True),
        tdataminsessions=prop('tdataminsessions', dtype='int', required=True),
        tdatadriver=prop('tdatadriver', dtype='string', required=True),
        tdatainsertonly=prop('tdatainsertonly', dtype='boolean', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        rmretdel=prop('rmretdel', dtype='boolean'),
        tdatabatchperiod=prop('tdatabatchperiod', dtype='int'),
        stage1tablename=prop('stage1tablename', dtype='string'),
        stage2tablename=prop('stage2tablename', dtype='string'),
        connectstring=prop('connectstring', dtype='string'),
        tdatatracelevel=prop('tdatatracelevel', dtype='int'),
        configfilesection=prop('configfilesection', dtype='string'),
        tdatauserpwdencrypted=prop('tdatauserpwdencrypted', dtype='boolean')
    )

    def __init__(self, tdatatdpid=None, desttablename=None, tdatausername=None,
                 tdatauserpwd=None, tdataminsessions=None, tdatamaxsessions=None,
                 tdatadriver=None, tdatainsertonly=None,
                 name=None, is_active=None, snapshot=None,
                 rmretdel=None, tdatabatchperiod=None, stage1tablename=None,
                 stage2tablename=None, connectstring=None, tdatatracelevel=None,
                 configfilesection=None, tdatauserpwdencrypted=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'tdata', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['tdatatdpid',
                                                   'desttablename',
                                                   'tdatausername',
                                                   'tdatauserpwd',
                                                   'tdataminsessions',
                                                   'tdatamaxsessions',
                                                   'tdatadriver',
                                                   'tdatainsertonly'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4], req[5], req[6],
                   req[7], name=name, is_active=is_active, **properties)


class TeradataListenerSubscriber(Connector):
    '''
    Subscribe to Teradata Listener events

    Parameters
    ----------
    ingestUrl : string
        Specifies the URL for the Listener Ingest REST API (version 1)
    SSLCert : strirng
        Specifies the path to a file that contains SSL certificates securely
        connect to the Listener Ingest service. Listener uses TLS 1.2.
    sourceKey : string
        Specifies the Listener source secret key that identifies the
        Listener source feed to which Event Stream Processing sends data.
    ingestBlocksize : int, optional
        Specifies the maximum number of data rows to send in one Listener
        Ingest message. Matching the connector block size to the Source
        window block size is recommended. The default block size is 256.
    contentType : string, optional
        Specifies the format of the data sent from Event Stream Processing
        to Listener, either JSON or plaintext (comma-delimited).
        The default is JSON.
    ingestDelim : string, optional
        Specifies the character that delimits data rows in a multi-row
        message from Event Stream Processing to the Listener Ingest
        REST API. The delimiter must not be a JSON punctuation character.
        The default is a tilde (~).
    snapshot : boolean, optional
         Specifies whether to send snapshot data.

    Returns
    -------
    :class:`TeradataListenerSubscriber`

    '''
    connector_key = dict(cls='tdlistener', type='subscribe')

    property_defs = dict(
        ingestUrl=prop('ingestUrl', dtype='string', required=True),
        SSLCert=prop('SSLCert', dtype='string', required=True),
        sourceKey=prop('sourceKey', dtype='string', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        ingestBlocksize=prop('ingestBlocksize', dtype='int'),
        contentType=prop('contentType', dtype='string'),
        ingestDelim=prop('ingestDelim', dtype='string')
    )

    def __init__(self, ingestUrl, SSLCert, sourceKey, name=None,
                 is_active=None, snapshot=None, ingestBlocksize=None,
                 contentType=None, ingestDelim=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'tdlistener', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['ingestUrl',
                                                   'SSLCert',
                                                   'sourceKey'],
                                         delete='type')
        return cls(req[0], req[1], req[2], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import copy
import datetime
import esppy
import esppy.utils.xml as xml
import os
import pandas as pd
import shutil
import six
import sys
import tempfile
import time
import unittest
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestProject(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def test_metadata(self):
        metadata = dict(self.project.metadata.items())

        self.project.metadata.clear()

        self.assertEqual(set(self.project.metadata.keys()), set())

        # Set keys
        self.project.metadata['foo'] = 'bar'
        self.project.metadata['one'] = '2'

        self.assertEqual(self.project.metadata, {'foo': 'bar', 'one': '2'})
        self.assertEqual(self.project.metadata['foo'], 'bar')
        self.assertEqual(self.project.metadata['one'], '2')

        with self.assertRaises(KeyError):
            self.project.metadata['none']

        # Update keys
        self.project.metadata['foo'] = 'baz'

        self.assertEqual(self.project.metadata, {'foo': 'baz', 'one': '2'})

        # Delete keys
        del self.project.metadata['foo']

        self.assertEqual(self.project.metadata, {'one': '2'})
        self.assertEqual(self.project.metadata['one'], '2')

        with self.assertRaises(KeyError):
            self.project.metadata['foo']

        del self.project.metadata['one']

        self.assertEqual(self.project.metadata, {})
        self.assertEqual(list(self.project.metadata.keys()), [])
        self.assertEqual(list(self.project.metadata.values()), [])

        # Update any original metadata
        self.project.metadata.update(metadata)

        self.assertEqual(self.project.metadata, metadata)

    def test_queries_str(self):
        self.assertEqual(str(self.project.queries),
                        "{'contquery': ContinuousQuery(name='contquery', project='ESPUnitTestProjectSA')}")

    def test_queries_repr(self):
        self.assertEqual(repr(self.project.queries),
                        "{'contquery': ContinuousQuery(name='contquery', project='ESPUnitTestProjectSA')}")

    def test_queries_len(self):
        self.assertEqual(len(self.project.queries), 1)

    def test_queries_iter(self):
        for item in self.project.queries:
            self.assertEqual(item, 'contquery')

    def test_get_windows(self):
        # Get all windows
        out = self.project.get_windows()
        self.assertEqual(len(out), 3)
        self.assertTrue('w_calculate' in out)
        self.assertTrue('w_data' in out)
        self.assertTrue('w_request' in out)

        # Subset names
        out = self.project.get_windows('w_data|w_request')
        self.assertEqual(len(out), 2)
        self.assertTrue('w_data' in out)
        self.assertTrue('w_request' in out)

        # Non-existent names
        out = self.project.get_windows('foo')
        self.assertEqual(len(out), 0)

        # Filters
        out = self.project.get_windows(filter='match(name, "data|request")')
        self.assertEqual(len(out), 2)
        self.assertTrue('w_data' in out)
        self.assertTrue('w_request' in out)

    def test_get_window(self):
        # Get specific window
        out = self.project.get_window('contquery.w_data')
        self.assertEqual(out.name, 'w_data')
        self.assertTrue(len(out.schema.fields) > 0)

        # Non-existent name
        with self.assertRaises(KeyError):
            self.project.get_window('contquery.foo')

    def test_name(self):
        self.assertEqual(self.project.name, 'ESPUnitTestProjectSA')

        for key, value in self.project.queries.items():
            self.assertEqual(value.project, 'ESPUnitTestProjectSA')
            for wkey, wvalue in value.windows.items():
                self.assertEqual(wvalue.project, 'ESPUnitTestProjectSA')

        self.project.name = 'Foo'

        self.assertEqual(self.project.name, 'Foo')

        for key, value in self.project.queries.items():
            self.assertEqual(value.project, 'Foo')
            for wkey, wvalue in value.windows.items():
                self.assertEqual(wvalue.project, 'Foo')

    def test_copy(self):
        # Use copy method directly
        proj = self.project.copy()
        self.assertEqual(proj.name, self.project.name)
        self.assertEqual(len(proj.queries), 1)
        self.assertEqual(sorted(proj.queries.keys()), ['contquery'])
        self.assertTrue(proj.queries['contquery'] is self.project.queries['contquery'])
        
        # Use copy.copy
        proj = copy.copy(self.project)
        self.assertEqual(proj.name, self.project.name)
        self.assertEqual(len(proj.queries), 1)
        self.assertEqual(sorted(proj.queries.keys()), ['contquery'])
        self.assertTrue(proj.queries['contquery'] is self.project.queries['contquery'])
        
        # Use copy(deep=True) method directly
        proj = self.project.copy(deep=True)
        self.assertEqual(proj.name, self.project.name)
        self.assertEqual(len(proj.queries), 1)
        self.assertEqual(sorted(proj.queries.keys()), ['contquery'])
        self.assertTrue(proj.queries['contquery'] is not self.project.queries['contquery'])
        self.assertEqual(sorted(proj.queries['contquery'].windows.keys()),
                         ['w_calculate', 'w_data', 'w_request'])
        self.assertTrue(proj['contquery']['w_calculate'] is not
                        self.project['contquery']['w_calculate'])
        self.assertTrue(proj['contquery']['w_data'] is not
                        self.project['contquery']['w_data'])
        self.assertTrue(proj['contquery']['w_request'] is not
                        self.project['contquery']['w_request'])

        # Use copy.deepcopy 
        proj = copy.deepcopy(self.project)
        self.assertEqual(proj.name, self.project.name)
        self.assertEqual(len(proj.queries), 1)
        self.assertEqual(sorted(proj.queries.keys()), ['contquery'])
        self.assertTrue(proj.queries['contquery'] is not self.project.queries['contquery'])
        self.assertEqual(sorted(proj.queries['contquery'].windows.keys()),
                         ['w_calculate', 'w_data', 'w_request'])
        self.assertTrue(proj['contquery']['w_calculate'] is not
                        self.project['contquery']['w_calculate'])
        self.assertTrue(proj['contquery']['w_data'] is not
                        self.project['contquery']['w_data'])
        self.assertTrue(proj['contquery']['w_request'] is not
                        self.project['contquery']['w_request'])

    def test_fullname(self):
        self.assertEqual(self.project.fullname, 'ESPUnitTestProjectSA')
        self.project.name = 'Foo'
        self.assertEqual(self.project.fullname, 'Foo')

    def test_from_xml(self):
        with self.assertRaises(ValueError):
             self.project.from_xml('<foo />')

    def test_to_element(self):
        before_xml = self.project.to_xml(pretty=True)
        after_xml = xml.to_xml(self.project.to_element(), pretty=True)
        self.assertEqual(before_xml, after_xml)

    def test_to_graph(self):
        # TODO: Not sure how to verify
        out = self.project.to_graph(schema=True)

    def test_repr_svg(self):
        # TODO: Not sure how to verify
        out = self.project._repr_svg_()

    def test_str(self):
        self.assertEqual(str(self.project), "Project(name='ESPUnitTestProjectSA')")

    def test_repr(self):
        self.assertEqual(repr(self.project), "Project(name='ESPUnitTestProjectSA')")

    def test_save(self):
        tmp = tempfile.mkdtemp()
        self.project.save(tmp)
        self.assertTrue(os.path.isdir(tmp))
        self.assertTrue(os.path.isfile(os.path.join(tmp, 'project.state')))
        self.assertTrue(os.path.isdir(os.path.join(tmp, 'queries')))
        shutil.rmtree(tmp, ignore_errors=True)

    def test_restore(self):
        tmp = tempfile.mkdtemp()
        self.project.save(tmp)
        self.project.stop()
        self.project.restore(tmp)
        self.assertTrue('ESPUnitTestProjectSA' in self.s.get_projects())
        shutil.rmtree(tmp, ignore_errors=True)

    @unittest.skip('Crashes server')
    def test_update(self):
        xml = self.project.to_xml()
        xml = xml.replace('name="w_data"', 'name="data_win"')
        xml = xml.replace('source="w_data"', 'source="data_win"')
        self.project.update(xml)

    def test_delete(self):
        self.assertIn('ESPUnitTestProjectSA', self.s.get_projects())
        self.project.delete()
        self.assertNotIn('ESPUnitTestProjectSA', self.s.get_projects())

    def test_create_mas_module(self):
        out = self.project.create_mas_module('python', 'mymod', ['func1', 'func2'])
        self.assertTrue(type(out).__name__, 'MASModule')

    def test_get_mas_modules(self):
        # TODO: Not sure what output should be
        out = self.project.get_mas_modules()

    @unittest.skip('Need a MAS module loaded')
    def test_get_mas_module(self):
        out = self.project.get_mas_modules(name)
        
    @unittest.skip('Need a MAS module loaded')
    def test_replace_mas_module(self):
        out = self.project.replace_mas_modules(name, module)

    def test_add_connectors(self):
        self.project.add_connectors('CG_1_sub',
                                    {'cq_01/tradesWindow/trades_sub': 'running'})
        self.project.add_connectors('CG_1_pub',
                                    {'cq_01/tradesWindow/trades_pub': 'finished'})

        elem = self.project.to_element()

        groups = {}
        for item in elem.findall('.//connector-groups/connector-group'):
            name = item.attrib['name']
            entries = {}
            for conn in item.findall('./connector-entry'):
                entries[conn.attrib['connector']] = conn.attrib['state'] 
            groups[name] = entries
        
        self.assertEqual(groups,
             {'CG_1_sub': {'cq_01/tradesWindow/trades_sub': 'running'},
              'CG_1_pub': {'cq_01/tradesWindow/trades_pub': 'finished'}})
        
    def test_get_stats(self):
        out = self.project.get_stats()
        self.assertTrue(isinstance(out.stats, pd.DataFrame))

    def test_validate(self):
        self.assertTrue(self.project.validate())

    def test_add_query(self):
        self.assertNotIn('new-query', self.project.queries)
        newq = self.project.add_query('new-query')
        self.assertIn('new-query', self.project.queries)

    def test_getitem(self):
        self.assertTrue(type(self.project['contquery']).__name__, 'ContinuousQuery')

        with self.assertRaises(KeyError):
            self.project['Foo']

    def test_setitem(self):
        with self.assertRaises(KeyError):
            self.project['new-query']

        newq = self.project.add_contquery('new-query')

        self.assertTrue(newq is self.project['new-query'])

        del self.project['new-query']

        with self.assertRaises(KeyError):
            self.project['new-query']

        self.project['new-query'] = newq

        self.assertTrue(newq is self.project['new-query'])

    def test_delitem(self):
        self.assertIn('contquery', self.project)
        del self.project['contquery']
        self.assertNotIn('contquery', self.project)

    def test_iter(self):
        for item in self.project:
            self.assertEqual(item, 'contquery')

    def test_len(self):
        self.assertEqual(len(self.project), 1)

    def test_contains(self):
        self.assertIn('contquery', self.project)
        self.assertNotIn('foo', self.project)

    def test_get_project_stats(self):
        ps = self.s.get_project_stats()
        self.assertEqual(len(ps), 0)


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Install the SAS ESP package '''

import glob
import io
import os
from setuptools import setup, find_packages

def get_file(fname):
    with io.open(os.path.join(os.path.dirname(os.path.abspath(__file__)), fname),
                 encoding='utf8') as infile:
        return infile.read()


setup(
    name='sas-esppy',
    version='6.2.2',
    description='SAS Event Stream Processing Python Interface',
    long_description=get_file('README.md'),
    long_description_content_type='text/markdown',
    author='SAS',
    author_email='Robert.Levey@sas.com',
    url='https://github.com/sassoftware/python-esppy/',
    license='Apache 2.0',
    packages=find_packages(),
    package_data={"esppy.templates": ["xmls/*.xml"]},
    install_requires=[
        'pandas >= 0.16.0',
        'pillow',
        'six >= 1.9.0',
        'plotly',
        'ipywidgets',
        'requests',
        # 'requests-kerberos',
        'graphviz',
        'rbtranslations',
        'ws4py',
        # 'wsaccel',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering',
    ],
)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Nurego Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class NuregoSubscriber(Connector):
    '''
    Subscribe to Nurego metering window

    Parameters
    ----------
    serviceurl : string
        Specifies the target Nurego REST service URL
    certificate : string
        Specifies the full path and filename of the client certificate that
        is used to establish the HTTPS connection to the Nurego REST service.
    username : string
        Specifies the user name to use in requests to Nurego for a new token
    password : string
        Specifies the password to use in requests to Nurego for a new token
    instanceid : string
        Specifies the instance ID to use in requests to Nurego for a new token
    certpassword : string, optional
        Specifies the password associated with the client certificate that
        is configured in certificate.
    collapse : string, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber output
        publishable. The default value is disabled.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks received by
        a subscriber that were introduced by a window retention policy.
    configfilesection : string, optional
        Specifies the name of the section in the config file to parse for
        configuration parameters. Specify the value as [configfilesection].

    Returns
    -------
    :class:`NuregoSubscriber`

    '''
    connector_key = dict(cls='nurego', type='subscribe')

    property_defs = dict(
        serviceurl=prop('serviceurl', dtype='string', required=True),
        certificate=prop('certificate', dtype='string', required=True),
        username=prop('username', dtype='string', required=True),
        password=prop('password', dtype='string', required=True),
        instanceid=prop('instanceid', dtype='string', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        certpassword=prop('certpassword', dtype='string'),
        collapse=prop('collapse', dtype='string'),
        rmretdel=prop('rmretdel', dtype='boolean'),
        configfilesection=prop('configfilesection', dtype='string')
    )

    def __init__(self, serviceurl=None, certificate=None, username=None,
                 password=None, instanceid=None,
                 name=None, is_active=None, snapshot=None,
                 certpassword=None, collapse=None, rmretdel=None,
                 configfilesection=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'nurego', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['serviceurl', 'certificate',
                                                   'username', 'password',
                                                   'instanceid'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4],
                   name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Rabbit MQ Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class RabbitMQSubscriber(Connector):
    '''
    Subscribe to Rabbit MQ events

    Parameters
    ----------
    rmqhost : string
        Specifies the Rabbit MQ server host name
    mqport : string
        Specifies the Rabbit MQ server port
    rmquserid : string
        Specifies the user name required to authenticate the connector's
        session with the Rabbit MQ server.    
    rmqpassword : string
        Specifies the password associated with rmquserid
    rmqexchange : string
        Specifies the Rabbit MQ exchange created by the connector,
        if nonexistent.
    rmqtopic : string
        Specifies the Rabbit MQ routing key to which messages are published.
    rmqtype : string
        Specifies binary, CSV, JSON, or the name of a string field in
        the subscribed window schema.
    urlhostport : string
        Specifies the host:port field in the metadata topic subscribed to
        on start-up to field metadata requests.
    numbufferedmsgs : int
        Specifies the maximum number of messages buffered by a standby
        subscriber connector.
    snapshot : boolean, optional
        Specifies whether to send snapshot data
    collapse : string, optional
        Enables conversion of UPDATE_BLOCK events to make subscriber
        output publishable. The default value is disabled.
    rmretdel : boolean, optional
        Specifies to remove all delete events from event blocks received
        by a subscriber that were introduced by a window retention policy.
    hotfailover : boolean, optional
        Enables hot failover mode
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields in
        CSV events. The default behavior is these fields are interpreted
        as an integer number of seconds (ESP_DATETIME) or microseconds
        (ESP_TIMESTAMP) since epoch.
    buspersistence : boolean, optional
        Specify to send messages using persistent delivery mode
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol Buffers
        message definition used to convert event blocks to protobuf messages.
        When you specify this parameter, you must also specify the
        protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in
        the .proto file that you specified with the protofile parameter.
        Event blocks are converted into this message.
    csvincludeschema : string, optional
        When rmqtype=CSV, specifies when to prepend output CSV data with
        the window's serialized schema. Valid values are never, once, and
        pereventblock. The default value is never.
    useclientmsgid : boolean, optional
        When performing a failover operation and extracting a message ID
        from an event block, use the client-generated message ID instead
        of the engine-generated message ID.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    rmqpasswordencrypted : boolean, optional
        Specifies that rmqpassword is encrypted
    rmqvhost : string, optional
        Specifies the Rabbit MQ vhost. The default is "/".
    csvmsgperevent : int, optional
        For CSV, specifies to send one message per event. The default is
        one message per transactional event block or else one message per event.
    csvmsgpereventblock : int, optional
        For CSV, specifies to send one message per event block. The default
        is one message per transactional event block or else one message
        per event.
    rmqcontenttype : string, optional
        Specifies the value of the content_type parameter in messages
        sent to RabbitMQ.
    rmqheaders : string, optional
        A comma separated list of key value optional headers in messages
        sent to RabbitMQ. The default value is no headers.
    rmqssl : boolean, optional
        Specifies to enable SSL encryption on the connection to the
        Rabbit MQ server.
    rmqsslcacert : string, optional
        When rmqssl is enabled, specifies the full path of the SSL CA
        certificate .pem file.
    rmqsslkey : string, optional
        When rmqssl is enabled, specifies the full path of the SSL key
        .pem file.
    rmqsslcert string, optional
        When rmqssl is enabled, specifies the full path of the SSL
        certificate .pem file.

    Returns
    -------
    :class:`RabbitMQSubscriber`

    '''
    connector_key = dict(cls='rmq', type='subscribe')

    property_defs = dict(
        rmquserid=prop('rmquserid', dtype='string', required=True),
        rmqpassword=prop('rmqpassword', dtype='string', required=True),
        rmqhost=prop('rmqhost', dtype='string', required=True),
        rmqport=prop('rmqport', dtype='int', required=True),
        rmqexchange=prop('rmqexchange', dtype='string', required=True),
        rmqtopic=prop('rmqtopic', dtype='string', required=True),
        rmqtype=prop('rmqtype', dtype='string', required=True),
        urlhostport=prop('urlhostport', dtype='string', required=True),
        numbufferedmsgs=prop('numbufferedmsgs', dtype='int', required=True),
        snapshot=prop('snapshot', dtype='boolean', required=True, default=False),
        collapse=prop('collapse', dtype='string'),
        rmretdel=prop('rmretdel', dtype='boolean'),
        hotfailover=prop('hotfailover', dtype='boolean'),
        dateformat=prop('dateformat', dtype='string'),
        buspersistence=prop('buspersistence', dtype='boolean'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        csvincludeschema=prop('csvincludeschema', dtype='string'),
        useclientmsgid=prop('useclientmsgid', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        rmqpasswordencrypted=prop('rmqpasswordencrypted', dtype='boolean'),
        rmqvhost=prop('rmqvhost', dtype='string'),
        csvmsgperevent=prop('csvmsgperevent', dtype='int'),
        csvmsgpereventblock=prop('csvmsgpereventblock', dtype='int'),
        rmqcontenttype=prop('rmqcontenttype', dtype='string'),
        rmqheaders=prop('rmqheaders', dtype='string'),
        rmqssl=prop('rmqssl', dtype='boolean'),
        rmqsslcacert=prop('rmqsslcacert', dtype='string'),
        rmqsslkey=prop('rmqsslkey', dtype='string'),
        rmqsslcert=prop('rmqsslcert', dtype='string')
    )

    def __init__(self, rmqhost=None, rmqport=None, rmquserid=None,
                 rmqpassword=None, rmqexchange=None,
                 rmqtopic=None, rmqtype=None, urlhostport=None, numbufferedmsgs=None,
                 snapshot=None, name=None, is_active=None, collapse=None,
                 rmretdel=None, hotfailover=None, dateformat=None,
                 buspersistence=None, protofile=None, protomsg=None,
                 csvincludeschema=None, useclientmsgid=None,
                 configfilesection=None, rmqpasswordencrypted=None,
                 rmqvhost=None, csvmsgperevent=None, csvmsgpereventblock=None,
                 rmqcontenttype=None, rmqheaders=None, rmqssl=None,
                 rmqsslcacert=None, rmqsslkey=None, rmqsslcert=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'rmq', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['rmqhost', 'rmqport', 'rmquserid',
                                                   'rmqpassword', 'rmqexchange',
                                                   'rmqtopic', 'rmqtype',
                                                   'urlhostport'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4], req[5], req[6],
                   req[7], name=name, is_active=is_active, **properties)


class RabbitMQPublisher(Connector):
    '''
    Publish Rabbit MQ events

    Parameters
    ----------
    rmqhost : string
        Specifies the Rabbit MQ server host name
    mqport : string
        Specifies the Rabbit MQ server port
    rmquserid : string
        Specifies the user name required to authenticate the connector’s
        session with the Rabbit MQ server.
    rmqpassword : string
        Specifies the password associated with rmquserid
    rmqexchange : string
        Specifies the Rabbit MQ exchange created by the connector,
        if nonexistent.
    rmqtopic : string
        Specifies the Rabbit MQ routing key to which messages are published.
    rmqtype : string
        Specifies binary, CSV, JSON, or the name of a string field in
        the subscribed window schema.
    urlhostport : string
        Specifies the host:port field in the metadata topic subscribed to
        on start-up to field metadata requests.
    transactional : string, optional
        When rmqtype=CSV, sets the event block type to transactional.
        The default value is normal.
    blocksize : int, optional
        When rmqtype=CSV, specifies the number of events to include in a
        published event block. The default value is 1.
    dateformat : string, optional
        Specifies the format of ESP_DATETIME and ESP_TIMESTAMP fields in
        CSV events. The default behavior is these fields are interpreted
        as an integer number of seconds (ESP_DATETIME) or microseconds
        (ESP_TIMESTAMP) since epoch.
    buspersistence : string, optional
        Controls both auto-delete and durable
    buspersistencequeue : string, optional
        Specifies the queue name used by a persistent publisher
    ignorecsvparseerrors : boolean, optional
        Specifies that when a field in an input CSV event cannot be
        parsed, the event is dropped, an error is logged, and
        publishing continues.
    protofile : string, optional
        Specifies the .proto file that contains the Google Protocol
        Buffers message definition used to convert event blocks to
        protobuf messages. When you specify this parameter, you must
        also specify the protomsg parameter.
    protomsg : string, optional
        Specifies the name of a Google Protocol Buffers message in the
        .proto file that you specified with the protofile parameter.
        Event blocks are converted into this message.
    configfilesection : string, optional
        Specifies the name of the section in the connector config file
        to parse for configuration parameters. Specify the value
        as [configfilesection].
    csvfielddelimiter : string, optional
        Specifies the character delimiter for field data in input CSV
        events. The default delimiter is the , character.
    noautogenfield : boolean, optional
        Specifies that input events are missing the key field that is
        autogenerated by the source window.
    ackwindow : int, optional
        Specifies the time period (in seconds) to leave messages that
        are received from Rabbit MQ unacknowledged. 
    acktimer : int, optional
        Specifies the time interval (in seconds) for how often to
        check whether to send acknowledgments that are triggered by
        the ackwindow parameter. Must be configured if ackwindow is
        configured.
    publishwithupsert : boolean, optional
        Builds events with opcode=Upsert instead of Insert
    rmqpasswordencrypted : boolean, optional
        Specifies that rmqpassword is encrypted
    addcsvopcode : boolean, optional
        Prepends an opcode and comma to input CSV events. The opcode
        is Insert unless publishwithupsert is enabled.
    addcsvflags : string, optional
        Specifies the event type to insert into input CSV events
        (with a comma). Valid values are "normal" and "partialupdate".
    rmqvhost : string, optional
        Specifies the Rabbit MQ vhost. The default is "/"
    useclientmsgid : string, optional
        If the Source window has been restored from a persist to disk,
        ignores received binary event blocks that contain a message ID
        less than the greatest message ID in the restored window.
    rmqssl : boolean, optional
        Specifies to enable SSL encryption on the connection to the
        Rabbit MQ server.
    rmqsslcacert : string, optional
        When rmqssl is enabled, specifies the full path of the SSL CA
        certificate .pem file.
    rmqsslkey : string, optional
        When rmqssl is enabled, specifies the full path of the SSL key
        .pem file.
    rmqsslcert string, optional
        When rmqssl is enabled, specifies the full path of the SSL
        certificate .pem file.
    maxevents : int, optional
         Specifies the maximum number of events to publish.

    Returns
    -------
    :class:`RabbitMQPublisher`

    '''
    connector_key = dict(cls='rmq', type='publish')

    property_defs = dict(
        rmquserid=prop('rmquserid', dtype='string', required=True),
        rmqpassword=prop('rmqpassword', dtype='string', required=True),
        rmqhost=prop('rmqhost', dtype='string', required=True),
        rmqport=prop('rmqport', dtype='int', required=True),
        rmqexchange=prop('rmqexchange', dtype='string', required=True),
        rmqtopic=prop('rmqtopic', dtype='string', required=True),
        rmqtype=prop('rmqtype', dtype='string', required=True),
        urlhostport=prop('urlhostport', dtype='string', required=True),
        transactional=prop('transactional', dtype='string'),
        blocksize=prop('blocksize', dtype='int'),
        dateformat=prop('dateformat', dtype='string'),
        buspersistence=prop('buspersistence', dtype='string'),
        buspersistencequeue=prop('buspersistencequeue', dtype='string'),
        ignorecsvparseerrors=prop('ignorecsvparseerrors', dtype='boolean'),
        protofile=prop('protofile', dtype='string'),
        protomsg=prop('protomsg', dtype='string'),
        configfilesection=prop('configfilesection', dtype='string'),
        csvfielddelimiter=prop('csvfielddelimiter', dtype='string'),
        noautogenfield=prop('noautogenfield', dtype='boolean'),
        ackwindow=prop('ackwindow', dtype='int'),
        acktimer=prop('acktimer', dtype='int'),
        publishwithupsert=prop('publishwithupsert', dtype='boolean'),
        rmqpasswordencrypted=prop('rmqpasswordencrypted', dtype='boolean'),
        addcsvopcode=prop('addcsvopcode', dtype='boolean'),
        addcsvflags=prop('addcsvflags', dtype='string'),
        rmqvhost=prop('rmqvhost', dtype='string'),
        useclientmsgid=prop('useclientmsgid', dtype='boolean'),
        rmqssl=prop('rmqssl', dtype='boolean'),
        rmqsslcacert=prop('rmqsslcacert', dtype='string'),
        rmqsslkey=prop('rmqsslkey', dtype='string'),
        rmqsslcert=prop('rmqsslcert', dtype='string'),
        maxevents=prop('maxevents', dtype='int')
    )

    def __init__(self, rmqhost=None, rmqport=None, rmquserid=None,
                 rmqpassword=None, rmqexchange=None,
                 rmqtopic=None, rmqtype=None, urlhostport=None,
                 name=None, is_active=None,
                 transactional=None, blocksize=None, dateformat=None,
                 buspersistence=None, buspersistencequeue=None,
                 ignorecsvparseerrors=None, protofile=None, protomsg=None,
                 configfilesection=None, csvfielddelimiter=None,
                 noautogenfield=None, ackwindow=None, acktimer=None,
                 publishwithupsert=None, rmqpasswordencrypted=None,
                 addcsvopcode=None, addcsvflags=None, rmqvhost=None,
                 useclientmsgid=None, rmqssl=None, rmqsslcacert=None,
                 rmqsslkey=None, rmqsslcert=None, maxevents=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'rmq', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['rmqhost', 'rmqport',
                                                   'rmquserid', 'rmqpassword',
                                                   'rmqexchange', 'rmqtopic',
                                                   'rmqtype', 'urlhostport'],
                                         delete='type')
        return cls(req[0], req[1], req[2], req[3], req[4], req[5], req[6],
                   req[7], name=name, is_active=is_active, **properties)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
xdict

A dictionary that allows setting nested keys by period (.) delimited strings

'''

from __future__ import print_function, division, absolute_import, unicode_literals

import copy
import re
import six


def _is_compound_key(key, types=six.string_types + (six.text_type, six.binary_type)):
    '''
    Check for a compound key name

    Parameters
    ----------
    key : string
        The key name to check
    types : list of types, optional
        The types of object to check

    Returns
    -------
    True
        If the key is compound (i.e., contains a '.')
    False
        If the key is not compound

    '''
    return isinstance(key, types) and '.' in key


class xdict(dict):
    '''
    Nested dictionary that allows setting of nested keys using '.' delimited strings

    Keys with a '.' in them automatically get split into separate keys.
    Each '.' in a key represents another level of nesting in the resulting
    dictionary.

    Parameters
    ----------
    *args, **kwargs : Arbitrary arguments and keyword arguments
        Same arguments as `dict`

    Returns
    -------
    xdict object

    Examples
    --------
    >>> dct = xdict()
    >>> dct['a.b.c'] = 100
    {'a': {'b': {'c': 100}}}

    '''

    def __init__(self, *args, **kwargs):
        super(xdict, self).__init__()
        self.update(*args, **kwargs)

    def __dir__(self):
        if hasattr(self, '_dir') and self._dir:
            return list(self._dir)
        return super(xdict, self).__dir__()

    def set_dir_values(self, values):
        '''
        Set the valid values for keys to display in tab-completion

        Parameters
        ----------
        values : iterable
           The values to display

        '''
        super(xdict, self).__setattr__('_dir', values)

    def set_doc(self, docstring):
        ''' Set the docstring for the xdict '''
        super(xdict, self).__setattr__('__doc__', docstring)

    def __copy__(self):
        return type(self)(**self)

    def __deepcopy__(self, memo):
        out = type(self)()
        for key, value in six.iteritems(self):
            if isinstance(value, (dict, list, tuple, set)):
                value = copy.deepcopy(value)
            out[key] = value
        return out

    @classmethod
    def from_json(cls, jsonstr):
        '''
        Create an xdict object from a JSON string

        Parameters
        ----------
        jsonstr : string
           Valid JSON string that represents an object

        Returns
        -------
        xdict object

        '''
        import json
        out = cls()
        out.update(json.loads(jsonstr))
        return out

    def __setitem__(self, key, value):
        ''' Set a key/value pair in an xdict object '''
        if isinstance(value, dict) and not isinstance(value, type(self)):
            value = type(self)(value)
        if _is_compound_key(key):
            return self._xset(key, value)
        return super(xdict, self).__setitem__(key, value)

    def _xset(self, key, value):
        '''
        Set a key/value pair allowing nested levels in the key

        Parameters
        ----------
        key : any
           Key value, if it is a string delimited by periods (.), each
           period represents another level of nesting of xdict objects.
        value : any
           Data value

        Returns
        -------
        None

        '''
        if isinstance(value, dict) and not isinstance(value, type(self)):
            value = type(self)(value)
        if _is_compound_key(key):
            current, key = key.split('.', 1)
            if current not in self:
                self[current] = type(self)()
            return self[current]._xset(key, value)
        self[key] = value

    def setdefault(self, key, *default):
        ''' Return keyed value, or set it to `default` if missing '''
        if _is_compound_key(key):
            try:
                return self[key]
            except KeyError:
                if default:
                    default = default[0]
                    if isinstance(default, dict) and not isinstance(default, type(self)):
                        default = type(self)(default)
                else:
                    default = None
                self[key] = default
                return default
        return super(xdict, self).setdefault(key, *default)

    def __contains__(self, key):
        ''' Does the xdict contain `key`? '''
        if super(xdict, self).__contains__(key):
            return True
        return key in self.allkeys()

    has_key = __contains__

    def __getitem__(self, key):
        ''' Get value stored at `key` '''
        if _is_compound_key(key):
            return self._xget(key)
        return super(xdict, self).__getitem__(key)

    def _xget(self, key, *default):
        '''
        Return keyed value, or `default` if missing

        Parameters
        ----------
        key : any
           Key to look up
        *default : any
           Default value to return if key is missing

        Returns
        -------
        any

        '''
        if _is_compound_key(key):
            current, key = key.split('.', 1)
            try:
                return self[current]._xget(key)
            except KeyError:
                if default:
                    return default[0]
                raise KeyError(key)
        return self[key]

    def get(self, key, *default):
        ''' Return keyed value, or `default` if missing '''
        if _is_compound_key(key):
            return self._xget(key, *default)
        return super(xdict, self).get(key, *default)

    def __delitem__(self, key):
        ''' Deleted keyed item '''
        if _is_compound_key(key):
            return self._xdel(key)
        super(xdict, self).__delitem__(key)

    def _xdel(self, key):
        '''
        Delete keyed item

        Parameters
        ----------
        key : any
           Key to delete.  If it is a string that is period (.) delimited,
           each period represents another level of nesting of xdict objects.

        Returns
        -------
        None

        '''
        if _is_compound_key(key):
            current, key = key.split('.', 1)
            try:
                return self[current]._xdel(key)
            except KeyError:
                raise KeyError(key)
        del self[key]

    def pop(self, key, *default):
        ''' Remove and return value stored at `key` '''
        try:
            out = self[key]
            del self[key]
            return out
        except KeyError:
            if default:
                return default[0]
            raise KeyError(key)

    def _flatten(self, dct, output, prefix=''):
        '''
        Create a new dict with keys flattened to period (.) delimited keys

        Parameters
        ----------
        dct : dict
           The dictionary to flatten
        output : dict
           The resulting dictionary (used internally in recursion)
        prefix : string
           Key prefix built from upper levels of nesting

        Returns
        -------
        dict

        '''
        if prefix:
            prefix = prefix + '.'
        for key, value in six.iteritems(dct):
            if isinstance(value, dict):
                if isinstance(key, int):
                    intkey = '%s[%s]' % (re.sub(r'\.$', r'', prefix), key)
                    self._flatten(value, prefix=intkey, output=output)
                else:
                    self._flatten(value, prefix=prefix + key, output=output)
            else:
                if isinstance(key, int):
                    intkey = '%s[%s]' % (re.sub(r'\.$', r'', prefix), key)
                    output[intkey] = value
                else:
                    output[prefix + key] = value

    def flattened(self):
        ''' Return an xdict with keys flattened to period (.) delimited strings '''
        output = {}
        self._flatten(self, output)
        return output

    def allkeys(self):
        ''' Return a list of all possible keys (even sub-keys) in the xdict '''
        out = set()
        for key in self.flatkeys():
            out.add(key)
            while '.' in key:
                key = key.rsplit('.', 1)[0]
                out.add(key)
                if '[' in key:
                    out.add(re.sub(r'\[\d+\]', r'', key))
        return list(out)

    def flatkeys(self):
        ''' Return a list of flattened keys in the xdict '''
        return list(self.flattened().keys())

    def flatvalues(self):
        ''' Return a list of flattened values in the xdict '''
        return list(self.flattened().values())

    def flatitems(self):
        ''' Return tuples of flattened key/value pairs '''
        return list(self.flattened().items())

    def iterflatkeys(self):
        ''' Return iterator of flattened keys '''
        return six.iterkeys(self.flattened())

    def iterflatvalues(self):
        ''' Return iterator of flattened values '''
        return six.itervalues(self.flattened())

    def iterflatitems(self):
        ''' Return iterator of flattened items '''
        return six.iteritems(self.flattened())

    def viewflatkeys(self):
        ''' Return view of flattened keys '''
        return six.viewkeys(self.flattened())

    def viewflatvalues(self):
        ''' Return view of flattened values '''
        return six.viewvalues(self.flattened())

    def viewflatitems(self):
        ''' Return view of flattened items '''
        return six.viewitems(self.flattened())

    def update(self, *args, **kwargs):
        ''' Merge the key/value pairs into `self` '''
        for arg in args:
            if isinstance(arg, dict):
                for key, value in six.iteritems(arg):
                    self._xset(key, value)
            else:
                for key, value in arg:
                    self._xset(key, value)
        for key, value in six.iteritems(kwargs):
            self._xset(key, value)

    def to_json(self):
        '''
        Convert an xdict object to a JSON string

        Returns
        -------
        string

        '''
        import json
        return json.dumps(self)


class xadict(xdict):
    '''
    An xdict that also allows setting/getting/deleting keys as attributes

    '''

    getdoc = None
    trait_names = None

    def _getAttributeNames(self):
        ''' Block this from creating attributes '''
        return

    def __delattr__(self, key):
        ''' Delete the attribute stored at `key` '''
        if key.startswith('_') and key.endswith('_'):
            return super(xadict, self).__delattr__(key)
        del self[key]

    def __getattr__(self, key):
        ''' Get the attribute store at `key` '''
        if key.startswith('_') and key.endswith('_'):
            return super(xadict, self).__getattr__(key)
        try:
            return self[key]
        except KeyError:
            dct = type(self)()
            self[key] = dct
            return dct

    def __getitem__(self, key):
        ''' Get item of an integer creates a new dict '''
        if isinstance(key, int) and key not in self:
            out = type(self)()
            self[key] = out
            return out
        return super(xadict, self).__getitem__(key)

    def __setattr__(self, key, value):
        ''' Set the attribute stored at `key` '''
        if key.startswith('_') and key.endswith('_'):
            return super(xadict, self).__setattr__(key, value)
        self[key] = value



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

# NOTE: This test requires a running ESP server.  You must use an ~/.authinfo
#       file to specify your username and password.  The ESP host and port must
#       be specified using the ESPHOST and ESPPORT environment variables.
#       A specific protocol ('http' or 'https') can be set using
#       the ESPPROTOCOL environment variable.

import datetime
import os
import six
import esppy
import sys
import unittest
from . import utils as tm

USER, PASSWD = tm.get_user_pass()
HOST, PORT, PROTOCOL = tm.get_host_port_proto()
DATA_DIR = tm.get_data_dir()


class TestAlgorithm(tm.TestCase):

    def setUp(self):
        self.s = esppy.ESP(HOST, PORT, USER, PASSWD, protocol=PROTOCOL)

        model_xml_path = os.path.join(DATA_DIR, 'model_sa.xml')
        name = 'ESPUnitTestProjectSA'

        with self.assertRaises(KeyError):
            self.s.get_project(name)

        self.project = self.s.install_project(model_xml_path)
        self.query = self.project.queries['contquery']

    def tearDown(self):
        self.s.delete_projects(filter="match(name, '^.*UnitTest.*$')")

    def test_str(self):
        out = str(self.s.get_algorithm('calculate', 'Summary'))
        self.assertEqual(out.replace("u'", "'"),
                         "Algorithm(name='Summary', parameters=['windowLength'], input_map=['input'], output_map=['cssOut', 'cvOut', 'maxOut', 'meanOut', 'minOut', 'nOut', 'nmissOut', 'stdOut', 'stderrOut', 'sumOut', 'ussOut', 'varOut'])")

    def test_repr(self):
        out = repr(self.s.get_algorithm('calculate', 'Summary'))
        self.assertEqual(out.replace("u'", "'"),
                         "Algorithm(name='Summary', parameters=['windowLength'], input_map=['input'], output_map=['cssOut', 'cvOut', 'maxOut', 'meanOut', 'minOut', 'nOut', 'nmissOut', 'stdOut', 'stderrOut', 'sumOut', 'ussOut', 'varOut'])")

    def test_subclass_str(self):
        self.assertTrue(hasattr(self.s, 'calculate'))
        self.assertTrue(hasattr(self.s.calculate, 'Summary'))

        out = str(self.s.calculate.Summary())
        self.assertRegex(out, r"Summary\(name=u?'w_\w+', contquery=None, project=None\)")

    def test_subclass_repr(self):
        self.assertTrue(hasattr(self.s, 'calculate'))
        self.assertTrue(hasattr(self.s.calculate, 'Summary'))

        out = repr(self.s.calculate.Summary())
        self.assertRegex(out, r"Summary\(name=u?'w_\w+', contquery=None, project=None\)")


if __name__ == '__main__':
   tm.runtests()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import copy
import unittest
from . import utils as tm
from esppy.utils.xdict import xdict, xadict

out1 = {
    'a': {'one': 1},
    'b': {'two': 2},
    'c': {
        'three': {'nest': 3},
        'four': {
            'nest': {'double': 4}
        }
    }
}

flatout1 = {
    'a.one': 1,
    'b.two': 2,
    'c.three.nest': 3,
    'c.four.nest.double': 4
}

class TestXDict(tm.TestCase):

    def test_constructor(self):
        x = xdict(one=1, two=2, three=3)
        self.assertEqual(x, dict(one=1, two=2, three=3))

        x = xdict([('a.one', 1), ('b.two', 2), ('c.three.nest', 3)])
        self.assertEqual(x, {
            'a': {'one': 1},
            'b': {'two': 2},
            'c': {
                'three': {'nest': 3}
            }
        })

        x = xdict([('a.one', 1), ('b.two', 2), ('c.three.nest', 3), ('c.four.nest.double', 4)])
        self.assertEqual(x, out1)

        x = xdict(out1)
        self.assertEqual(x, out1)

        x = xdict(flatout1)
        self.assertEqual(x, out1)

        x = xdict(**out1)
        self.assertEqual(x, out1)
      
        x = xdict(**flatout1)
        self.assertEqual(x, out1)

    def test_setitem(self):
        x = xdict()
        x['a.one'] = 1
        x['b.two'] = 2
        x['c.three.nest'] = 3
        x['c.four.nest.double'] = 4
        self.assertEqual(x, out1)

        x[100] = 1000
        self.assertEqual(x[100], 1000)

        x = xdict()
        x['a.one'] = 1
        x['b.two'] = 2
        x['c.three.nest'] = 3
        x['c.four'] = {'nest': {'double': 4}}
        self.assertEqual(x, out1)

    def test_getitem(self):
        x = xdict(out1)
        self.assertEqual(x['a.one'], 1)
        self.assertEqual(x['b.two'], 2)
        self.assertEqual(x['c.three.nest'], 3)
        self.assertEqual(x['c.four.nest.double'], 4)

        self.assertEqual(x['a'], {'one': 1})
        self.assertEqual(x['b'], {'two': 2})
        self.assertEqual(x['c'], {'three': {'nest': 3}, 'four': {'nest': {'double': 4}}})
        self.assertEqual(x['c.three'], {'nest': 3})
        self.assertEqual(x['c.four'], {'nest': {'double': 4}})
        self.assertEqual(x['c.four.nest'], {'double': 4})
        
        self.assertEqual(x['a'], {'one': 1})
        self.assertEqual(x['a']['one'], 1)
        self.assertEqual(x['b'], {'two': 2})
        self.assertEqual(x['b']['two'], 2)
        self.assertEqual(x['c'], {'three': {'nest': 3}, 'four': {'nest': {'double': 4}}})
        self.assertEqual(x['c']['three'], {'nest': 3})
        self.assertEqual(x['c']['three']['nest'], 3)
        self.assertEqual(x['c']['four'], {'nest': {'double': 4}})
        self.assertEqual(x['c']['four']['nest'], {'double': 4})
        self.assertEqual(x['c']['four']['nest']['double'], 4)
        
    def test_get(self):
        x = xdict(out1)
        
        self.assertEqual(x.get('a', 1000), {'one': 1})
        self.assertEqual(x.get('a.one', 1000), 1)
        self.assertEqual(x.get('a.does.not.exist', 1000), 1000)

    def test_delitem(self):
        newout1 = copy.deepcopy(out1)
        x = xdict(out1)

        del x['a']
        del newout1['a']
        self.assertEqual(x, newout1)

        del x['c.four.nest']
        del newout1['c']['four']['nest']
        self.assertEqual(x, newout1)

        with self.assertRaises(KeyError):
            del x['a.does.not.exist']
        
    def test_flattened(self):
        x = xdict(out1)
        self.assertEqual(x.flattened(), flatout1)        

    def test_setdefault(self):
        x = xdict(out1)

        self.assertEqual(x.setdefault('a.one'), 1)
        self.assertEqual(x.setdefault('b.two'), 2)
        self.assertEqual(x.setdefault('b'), {'two': 2})

        self.assertEqual(x.setdefault('w.none'), None)
        self.assertEqual(x['w.none'], None)
        self.assertEqual(x['w']['none'], None)
        self.assertEqual(x.setdefault('x.y.z', 200), 200)
        self.assertEqual(x['x.y.z'], 200)
        self.assertEqual(x['x']['y']['z'], 200)
        self.assertEqual(x.setdefault('x.y.z', 10), 200)
        self.assertEqual(x['x.y.z'], 200)
        self.assertEqual(x['x']['y']['z'], 200)
        self.assertEqual(x['x']['y'], {'z': 200})

        self.assertEqual(x.setdefault('does.not.exist', {'new': 'key'}), {'new': 'key'})
        self.assertTrue(isinstance(x.setdefault('does.not.exist', {'new': 'key'}), xdict))

    def test_json(self):
        x = xdict(out1)
        self.assertEqual(xdict.from_json(x.to_json()), out1)

    def test_flat(self):
        x = xdict(out1)

        self.assertEqual(sorted(x.flatkeys()), sorted(flatout1.keys()))
        self.assertEqual(sorted(x.flatvalues()), sorted(flatout1.values()))
        self.assertEqual(sorted(x.flatitems()), sorted(flatout1.items()))

        if hasattr(flatout1, 'iterkeys'):
            self.assertEqual(sorted(x.iterflatkeys()), sorted(flatout1.iterkeys()))
        if hasattr(flatout1, 'itervalues'):
            self.assertEqual(sorted(x.iterflatvalues()), sorted(flatout1.itervalues()))
        if hasattr(flatout1, 'iteritems'):
            self.assertEqual(sorted(x.iterflatitems()), sorted(flatout1.iteritems()))

        if hasattr(flatout1, 'iterkeys'):
            self.assertEqual(sorted(x.viewflatkeys()), sorted(flatout1.iterkeys()))
        if hasattr(flatout1, 'itervalues'):
            self.assertEqual(sorted(x.viewflatvalues()), sorted(flatout1.itervalues()))
        if hasattr(flatout1, 'iteritems'):
            self.assertEqual(sorted(x.viewflatitems()), sorted(flatout1.iteritems()))

    def test_contains(self):
        x = xdict(out1)

        self.assertTrue('a' in x)
        self.assertTrue('a.one' in x)
        self.assertTrue('b' in x)
        self.assertTrue('c.four' in x)
        self.assertTrue('c.four.nest' in x)
        self.assertTrue('c.four.nest.double' in x)
        self.assertFalse('z' in x)
        self.assertFalse('four' in x)

        self.assertTrue(x.has_key('a'))
        self.assertTrue(x.has_key('a.one'))
        self.assertTrue(x.has_key('b'))
        self.assertTrue(x.has_key('c.four'))
        self.assertTrue(x.has_key('c.four.nest'))
        self.assertTrue(x.has_key('c.four.nest.double'))
        self.assertFalse(x.has_key('z'))
        self.assertFalse(x.has_key('four'))

    def test_pop(self):
        x = xdict(out1)

        self.assertEqual(x, out1)

        self.assertEqual(x.pop('a.one'), out1['a']['one'])
        self.assertEqual(x.pop('b.two'), out1['b']['two'])
        self.assertEqual(x.pop('c'), out1['c'])

        self.assertEqual(x, {'a': {}, 'b': {}})

        with self.assertRaises(KeyError):
            x.pop('a.does.not.exist')

        self.assertEqual(x.pop('a.does.not.exist', 1000), 1000)

    def test_copy(self):
        w = xdict(out1)

        x = w.copy()

        self.assertEqual(w, x)
        self.assertTrue(w is not x)
        self.assertTrue(w['a'] is x['a'])
        self.assertTrue(w['b'] is x['b'])
        self.assertTrue(w['c'] is x['c'])

        y = copy.copy(w)

        self.assertEqual(w, y)
        self.assertTrue(w is not y)
        self.assertTrue(w['a'] is y['a'])
        self.assertTrue(w['a.one'] is y['a.one'])
        self.assertTrue(w['b'] is y['b'])
        self.assertTrue(w['b.two'] is y['b.two'])
        self.assertTrue(w['c'] is y['c'])
        self.assertTrue(w['c.three'] is y['c.three'])

        z = copy.deepcopy(w)

        # TODO: Deepcopy is only shallow copying
        self.assertEqual(w, z)
        self.assertTrue(w is not z)
        self.assertTrue(w['a'] is not z['a'])
#       self.assertTrue(w['a.one'] is not z['a.one'])
        self.assertTrue(w['b'] is not z['b'])
#       self.assertTrue(w['b.two'] is not z['b.two'])
        self.assertTrue(w['c'] is not z['c'])
#       self.assertTrue(w['c.three'] is not z['c.three'])

    def test_attrs(self):
        x = xadict(out1)

        self.assertEqual(x.a, x['a'])
        self.assertEqual(x.a.one, x['a.one'])
        self.assertEqual(x.a.one, x['a']['one'])
        self.assertEqual(x.b, x['b'])
        self.assertEqual(x.b.two, x['b.two'])
        self.assertEqual(x.b.two, x['b']['two'])
        self.assertEqual(x.c, x['c'])
        self.assertEqual(x.c.three, x['c.three'])
        self.assertEqual(x.c.three, x['c']['three'])

        x.a.one = 100
        x.c.four.nest = {'float': 5}

        self.assertEqual(x['a.one'], 100)
        self.assertEqual(x['c.four.nest'], {'float': 5})

        del x.c.three

        self.assertEqual(x.c, {'four': {'nest': {'float': 5}}})

        del x.c

        self.assertEqual(x, {'a': {'one': 100}, 'b': {'two': 2}})

        del x.a
        del x.b

        self.assertEqual(x, {})

        # Non-existent keys
        x.foo.bar.xxx = 10
        x.foo.baz = 'hi there'

        self.assertEqual(x, {'foo': {'bar': {'xxx': 10}, 'baz': 'hi there'}})

    def test_ints(self):
        x = xadict(out1)

        out1int = dict(out1)
        out1int['d'] = {}
        out1int['d'][0] = dict(hi='there', bye='now')
        out1int['d'][1] = dict(test='value')
        out1int['d'][2] = 100

        x.d[0].hi = 'there'
        x.d[0].bye = 'now'
        x.d[1].test = 'value'
        x.d[2] = 100

        self.assertEqual(x, out1int)

        # Flat
        flatout1int = dict(flatout1) 
        flatout1int['d[0].hi'] = 'there'
        flatout1int['d[0].bye'] = 'now'
        flatout1int['d[1].test'] = 'value'
        flatout1int['d[2]'] = 100

        self.assertEqual(x.flattened(), flatout1int)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

'''
Utilities for reading authinfo/netrc files

'''

from __future__ import print_function, division, absolute_import, unicode_literals

import os
import re
import sys

_AUTHINFO_PATHS = [
    '_authinfo.gpg',
    '.authinfo.gpg',
    '_netrc.gpg',
    '.netrc.gpg',
    '_authinfo',
    '.authinfo',
    '_netrc',
    '.netrc',
]

if 'win' not in sys.platform.lower():
    _AUTHINFO_PATHS = [aipath for aipath in _AUTHINFO_PATHS if not aipath.startswith('_')]

_ALIASES = {
    'machine': 'host',
    'login': 'user',
    'account': 'user',
    'port': 'protocol',
}


def _chunker(seq, size):
    ''' Read sequence `seq` in `size` sized chunks '''
    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


def _matches(params, **kwargs):
    ''' See if keyword arguments are a subset of `params` '''
    for key, value in kwargs.items():
        if value is None:
            continue
        if key not in params:
            continue
        if params.get(key) != value:
            return False
    return True


def parseparams(param):
    '''
    Parse the next parameter from the string

    Parameters
    ----------
    param : string
        The string to parse

    Returns
    -------
    dict
        Key/value pairs parsed from the string

    '''
    out = {}

    if not param:
        return out

    siter = iter(param)

    name = []
    for char in siter:
        if not char.strip():
            break
        name.append(char)

    value = []
    for char in siter:
        if not char.strip():
            break
        if char == '"':
            for subchar in siter:
                if subchar == '\\':
                    value.append(next(siter))
                elif subchar == '"':
                    break
        else:
            value.append(char)

    name = ''.join(name)
    value = ''.join(value)

    out[_ALIASES.get(name, name)] = value
    out.update(parseparams((''.join(list(siter))).strip()))
    return out


def query_authinfo(host, user=None, protocol=None, path=None):
    '''
    Look for a matching host definition in authinfo/netrc files

    Parameters
    ----------
    host : string
        The host name or IP address to match
    user : string, optional
        The username to match
    protocol : string or int, optional
        The protocol or port to match
    path : string or list of strings, optional
        The paths to look for instead of the automatically detected paths

    Returns
    -------
    dict
        Connection information

    '''
    paths = []

    # Construct list of authinfo/netrc paths
    if path is None:
        if os.environ.get('AUTHINFO'):
            paths = [os.path.expanduser(x)
                     for x in os.environ.get('AUTHINFO').split(os.path.sep)]
        elif os.environ.get('NETRC'):
            paths = [os.path.expanduser(x)
                     for x in os.environ.get('NETRC').split(os.path.sep)]
        else:
            home = os.path.expanduser('~')
            for item in _AUTHINFO_PATHS:
                paths.append(os.path.join(home, item))

    elif not isinstance(path, (tuple, list, set)):
        paths = [os.path.expanduser(path)]

    else:
        paths = [os.path.expanduser(x) for x in path]

    # Parse each file
    for path in paths:

        if not os.path.exists(path):
            continue

        # Remove comments and macros
        lines = []

        try:
            with open(path) as info:
                infoiter = iter(info)
                for line in infoiter:
                    line = line.strip()

                    # Bypass comments
                    if line.startswith('#'):
                        continue

                    # Bypass macro definitions
                    if line.startswith('macdef'):
                        for line in infoiter:
                            if not line.strip():
                                break
                        continue

                    lines.append(line)

        except OSError:
            continue

        line = ' '.join(lines)

        # Parse out definitions and look for matches
        defs = [x for x in re.split(r'\b(host|machine|default)\b\s*', line) if x.strip()]

        for name, value in _chunker(defs, 2):
            if name in ['host', 'machine']:
                hostname, value = re.split(r'\s+', value, 1)
                out = parseparams(value)
                out['host'] = hostname.lower()
                if _matches(out, host=host.lower(), user=user, protocol=protocol):
                    return out
            else:
                out = parseparams(value)
                if _matches(out, user=user, protocol=protocol):
                    return out

    return {}



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from __future__ import print_function, division, absolute_import, unicode_literals

from .base import BaseWindow, Window, get_window_class, Target
from .subscriber import Subscriber
from .publisher import Publisher
from .aggregate import AggregateWindow
from .calculate import CalculateWindow
from .copy import CopyWindow
from .compute import ComputeWindow
from .counter import CounterWindow
from .filter import FilterWindow
from .functional import FunctionalWindow
from .geofence import GeofenceWindow
from .join import JoinWindow
from .modelsuper import ModelSupervisorWindow
from .modelreader import ModelReaderWindow
from .notification import NotificationWindow
from .objectTracker import ObjectTrackerWindow
from .pattern import PatternWindow
from .procedural import ProceduralWindow
from .removeState import RemoveStateWindow
from .score import ScoreWindow
from .source import SourceWindow
from .textcategory import TextCategoryWindow
from .textcontext import TextContextWindow
from .textsentiment import TextSentimentWindow
from .texttopic import TextTopicWindow
from .train import TrainWindow
from .transpose import TransposeWindow
from .union import UnionWindow
from .pythonmas import PythonHelper


def get_subclasses(cls):
    for subclass in cls.__subclasses__():
        for subcls in get_subclasses(subclass):
            yield subcls
        yield subclass


for cls in get_subclasses(BaseWindow):
    if cls.window_type and not cls.is_hidden:
        BaseWindow.window_classes['window-%s' % cls.window_type] = cls



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Window Publisher '''

from __future__ import print_function, division, absolute_import, unicode_literals

import collections
import copy
import csv
import datetime
import functools
import itertools
import os
import pandas as pd
import re
import requests
import six
import sys
import threading
import types
import weakref
import xml.etree.ElementTree as ET
from six.moves import urllib
from .utils import verify_window
from ..utils.authorization import Authorization
from ..base import ESPObject, attribute
from ..config import get_option
from ..exceptions import ESPError
from ..plotting import StreamingChart, StreamingImages, split_chart_params
from ..schema import Schema
from ..utils.keyword import dekeywordify
from ..utils import xml
from ..utils.notebook import scale_svg
from ..utils.rest import get_params
from ..utils.data import get_project_data, gen_name, get_server_info
from ..utils.events import get_events, get_dataframe, get_schema
from ..websocket import WebSocketClient


class Publisher(object):
    '''
    Create a publisher for the given window

    Attributes
    ----------
    blocksize : int
        Number of events to put into an event block
    dateformat : string
        Format for date fields
    format : string
        The data format of inputs: 'csv', 'xml', 'json', 'properties'
    is_active : bool
        Is the web socket currently active?
    opcode : string
        Opcode to use if an input event does not include one:
        'insert', 'upsert', 'delete'
    pause : int
        Number of milliseconds to pause between each injection of events
    rate : int
        Maximum number of events to inject per second
    separator : string
        The separator string to use between events in 'properties' format
    window_schema : Schema
        The schema of the window that was subscribed to
    window_url : string
        The publisher URL of the window

    Parameters
    ----------
    window : Window
        The window object to create the subscriber for
    blocksize : int, optional
        Number of events to put into an event block
    rate : int, optional
        Maximum number of events to inject per second
    pause : int, optional
        Number of milliseconds to pause between each injection of events
    dateformat : string, optional
        Format for date fields
    opcode : string, optional
        Opcode to use if an input event does not include one:
        'insert', 'upsert', 'delete'
    format : string, optional
        The data format of inputs: 'csv', 'xml', 'json', 'properties'
    separator : string
        The separator string to use between events in 'properties' format

    Examples
    --------
    Create the publisher instance using CSV and an event rate of 200
    events per millisecond.

    >>> pub = Publisher(window, rate=200)

    Send the CSV data.

    >>> pub.send('1,2,3')

    Close the connection.

    >>> pub.close()

    Returns
    -------
    :class:`Publisher`

    '''

    def __init__(self, window, blocksize=1, rate=0, pause=0,
                 dateformat='%Y%m%dT%H:%M:%S.%f', opcode='insert',
                 format='csv', separator=None):
        self.blocksize = int(blocksize)
        self.rate = int(rate)
        self.pause = int(pause)
        self.dateformat = dateformat
        self.opcode = opcode
        self.format = format
        self.separator = separator
        self.session = window.session
        self.window_fullname = window.fullname
        self.window_schema = get_schema(window, window)
        self.window_url = window.publisher_url

        if not verify_window(window):
            raise ESPError('There is no window at %s' % window.fullname)

        if get_option('debug.requests'):
            sys.stderr.write('WEBSOCKET %s\n' % self.url)

        headers = []

        auth = Authorization.getInstance(self.session)

        if auth.isEnabled:
            headers.append(("Authorization",auth.authorization));

        self._ws = WebSocketClient(self.url,headers=headers)
        self._ws.connect()

    @property
    def url(self):
        '''
        Return the URL of the subscriber

        Returns
        -------
        string

        '''
        url_params = get_params(**{'rate': self.rate,
                                   'blocksize': self.blocksize,
                                   'pause': self.pause,
                                   'format': self.format,
                                   'dateformat': self.dateformat,
                                   'opcode': self.opcode,
                                   'separator': self.separator})
        url_params = urllib.parse.urlencode(sorted(url_params.items()))
        return self.window_url + '?' + url_params.replace('+', '%20')

    @property
    def is_active(self):
        '''
        Is the web socket active?

        Returns
        -------
        bool

        '''
        return self._ws is not None

    def send(self, data):
        '''
        Send data to the web socket

        Examples
        --------
        Create the publisher instance using CSV and an event rate of 200
        events per millisecond.

        >>> pub = Publisher(window, rate=200)

        Send the CSV data.

        >>> pub.send('1,2,3')

        Parameters
        ----------
        data : string
            The data to send

        '''
        if self._ws is None:
            raise ValueError('The connection is closed')
        return self._ws.send(data)

    def close(self):
        '''
        Close the web socket connection

        Examples
        --------
        Create the publisher instance using CSV and an event rate of 200
        events per millisecond.

        >>> pub = Publisher(window, rate=200)

        Send the CSV data.

        >>> pub.send('1,2,3')

        Close the connection.

        >>> pub.close()

        '''
        if self._ws is not None:
            self._ws.close()
            self._ws = None



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' ESP Adapter Connector '''

from __future__ import print_function, division, absolute_import, unicode_literals

import numbers
import re
import six
from .base import Connector, prop, map_properties
from ..utils import xml
from ..utils.data import gen_name


class AdapterSubscriber(Connector):
    '''
    Subscribe using an adapter

    Parameters
    ----------
    command : string
        Specifies the command and options that are used to run the adapter
        from the command line. All required adapter parameters must be specified
        when defining the command parameter.
    url : string, optional
        Specifies the URL for the adapter connection. If this parameter is not
        included, a default URL is generated.

    Returns
    -------
    :class:`AdapterSubscriber`

    '''
    connector_key = dict(cls='adapter', type='subscribe')

    property_defs = dict(
        command=prop('command', dtype='string', required=True),
        url=prop('url', dtype='string'),
    )

    def __init__(self, command, url=None, name=None, is_active=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'adapter', name=name, type='subscribe',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['command'],
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)


class AdapterPublisher(Connector):
    '''
    Publish using an adapter

    Parameters
    ----------
    command : string
        Specifies the command and options that are used to run the adapter
        from the command line. All required adapter parameters must be specified
        when defining the command parameter.
    url : string, optional
        Specifies the URL for the adapter connection. If this parameter is not
        included, a default URL is generated.

    Returns
    -------
    :class:`AdapterPublisher`

    '''
    connector_key = dict(cls='adapter', type='publish')

    property_defs = dict(
        command=prop('command', dtype='string', required=True),
        url=prop('url', dtype='string'),
    )

    def __init__(self, command, url=None, name=None, is_active=None):
        params = dict(**locals())
        params.pop('is_active')
        params.pop('self')
        name = params.pop('name')
        Connector.__init__(self, 'adapter', name=name, type='publish',
                           is_active=is_active, properties=params)

    @classmethod
    def from_parameters(cls, conncls, type=None, name=None, is_active=None,
                        properties=None):
        req, properties = map_properties(cls, properties,
                                         required=['command'],
                                         delete='type')
        return cls(req[0], name=name, is_active=is_active, **properties)
